"""
..
    Copyright (c) 2015-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing functions for calculating the updated value of mu (the
relative step-size used in the gradient descent iteration)  used in the
Smoothed l0 algorithms.

Routine listings
----------------
calculate_using_fixed(var)
    Calculate the updated, fixed mu value.
calculate_using_step(var)
    Calculate an updated mu value in a 'step' way.
get_function_handle(method)
    Return a function handle to a given calculation method.

"""

from __future__ import division


def wrap_calculate_using_fixed(var):
    """
    Arguments wrapper for calculate_using_fixed.

    """

    convert = var['convert']
    mu = convert(var['param']['mu_fixed'])

    def calculate_using_fixed(var):
        """
        Calculate the updated, fixed mu value.

        Parameters
        ----------
        var : dict
            Dictionary of variables used in the calculation of the updated mu
            value.

        Returns
        -------
        mu : float
            The updated mu value to be used in the SL0 algorithm.

        """

        return mu

    return calculate_using_fixed


def wrap_calculate_using_step(var):
    """
    Arguments wrapper for calculate_using_step.

    """

    convert = var['convert']
    sigma_start = var['sigma_start']
    sigma_geometric = var['sigma_geometric']
    mu_start = convert(var['param']['mu_step_start'])
    mu_end = convert(var['param']['mu_step_end'])
    mu_iteration = var['param']['mu_step_iteration']

    def calculate_using_step(var):
        """
        Calculate an updated mu value in a 'step' way.

        Parameters
        ----------
        var : dict
            Dictionary of variables used in the calculation of the updated mu
            value.

        Returns
        -------
        mu : float
            The updated mu value to be used in the SL0 algorithm.

        """

        i = var['i']

        if i < mu_iteration or sigma_start * sigma_geometric**i > 0.75:
            mu = mu_start
        else:
            mu = mu_end

        return mu

    return calculate_using_step


def get_function_handle(method, var):
    """
    Return a function handle to a given calculation method.

    Parameters
    ----------
    method : str
        Identifier of the calculation method to return a handle to.
    var : dict
        Local variables needed in the mu update method.

    Returns
    -------
    f_handle : function
        Handle to the calculation method defined in this globals scope.

    """

    return globals()['wrap_calculate_using_' + method](var)



"""
..
    Copyright (c) 2015-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing the core Generalised Approximate Message Passing (GAMP)
algorithm.

Routine listings
----------------
run(y, A, A_asq=None)
    Run the GAMP reconstruction algorithm.

See Also
--------
magni.cs.reconstruction.gamp._config : Configuration options.
magni.cs.reconstruction.gamp.input_channel : Available input channels.
magni.cs.reconstruction.gamp.output_channel : Available output channels.
magni.cs.reconstruction.gamp.stop_criterion : Available stop critria.

Notes
-----
The default configuration of the GAMP algorithm provides the s-GB AMP algorithm
from [1]_ using an MSE convergence based stop criterion. Both the input
channel, the output channel, and the stop criterion may be changed.

This implementation allows for the use of sum approximations of the squared
system matrix as detailed in [1]_ and [2]_. Furthermore, a simple damping
option is available based on the description in [3]_ (see also [4]_ for more
details on damping in GAMP).

References
----------
.. [1] F. Krzakala, M. Mezard, F. Sausset, Y. Sun, and L. Zdeborova,
   "Probabilistic reconstruction in compressed sensing: algorithms, phase
   diagrams, and threshold achieving matrices", *Journal of Statistical
   Mechanics: Theory and Experiment*, vol. P08009, pp. 1-57, Aug. 2012.
.. [2] S. Rangan, "Generalized Approximate Message Passing for Estimation
   with Random Linear Mixing", arXiv:1010.5141v2, pp. 1-22, Aug. 2012.
.. [3] S. Rangan, P. Schniter, and A. Fletcher. "On the Convergence of
   Approximate Message Passing with Arbitrary Matrices", *in IEEE International
   Symposium on Information Theory (ISIT)*, pp. 236-240, Honolulu, Hawaii, USA,
   Jun. 29 - Jul. 4, 2014.
.. [4] J. Vila, P. Schniter, S. Rangan, F. Krzakala, L. Zdeborova, "Adaptive
   Damping and Mean Removal for the Generalized Approximate Message Passing
   Algorithm", *in IEEE International Conference on Acoustics, Speech, and
   Signal Processing (ICASSP)*, South Brisbane, Queensland, Australia, Apr.
   19-24, 2015, pp. 2021-2025.

"""

from __future__ import division
import copy

import numpy as np

from magni.cs.reconstruction.gamp import config as _conf
from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate_generic as _generic
from magni.utils.validation import validate_numeric as _numeric
from magni.utils.validation.types import MatrixBase as _MatrixBase
from magni.utils.matrices import norm as _norm
from magni.utils.matrices import (SumApproximationMatrix as
                                  _SumApproximationMatrix)


def run(y, A, A_asq=None):
    """
    Run the GAMP reconstruction algorithm.

    Parameters
    ----------
    y : ndarray
        The m x 1 measurement vector.
    A : ndarray or magni.utils.matrices.{Matrix, MatrixCollection}
        The m x n matrix which is the product of the measurement matrix and the
        dictionary matrix.
    A_asq : ndarray or magni.utils.matrices.{Matrix, MatrixCollection} or None
        The m x n matrix which is the entrywise absolute value squared product
        of the measurement matrix and the dictionary matrix (the default is
        None, which implies that a sum approximation is used).

    Returns
    -------
    alpha : ndarray
        The n x 1 reconstructed coefficient vector.
    history : dict, optional
        The dictionary of various measures tracked in the GAMP iterations.

    See Also
    --------
    magni.cs.reconstruction.gamp._config : Configuration options.
    magni.cs.reconstruction.gamp.input_channel : Input channels.
    magni.cs.reconstruction.gamp.output_channel : Output channels.
    magni.cs.reconstruction.gamp.stop_criterion : Stop criteria.

    Notes
    -----
    Optionally, the algorithm may be configured to save and return the
    iteration history along with the reconstruction result. The returned
    history contains the following:

    * alpha_bar : Mean coefficient estimates (the reconstruction coefficients).
    * alpha_tilde : Variance coefficient estimates.
    * MSE : solution Mean squared error (if the true solution is known).
    * input_channel_parameters : The state of the input channel.
    * output_channel_parameters : The state of the output channel.
    * stop_criterion : The currently used stop criterion.
    * stop_criterion_value : The value of the stop criterion.
    * stop_iteration : The iteration at which the algorithm stopped.
    * stop_reason : The reason for termination of the algorithm.

    Examples
    --------
    For example, recovering a vector from AWGN noisy measurements using GAMP

    >>> import numpy as np
    >>> from magni.cs.reconstruction.gamp import run, config
    >>> np.random.seed(seed=6028)
    >>> k, m, n = 10, 200, 400
    >>> tau = float(k) / n
    >>> A = 1 / np.sqrt(m) * np.random.randn(m, n)
    >>> A_asq = np.abs(A)**2
    >>> alpha = np.zeros((n, 1))
    >>> alpha[:k] = np.random.normal(scale=2, size=(k, 1))
    >>> np_printoptions = np.get_printoptions()
    >>> np.set_printoptions(suppress=True, threshold=k+2)
    >>> alpha[:k + 2]
    array([[ 1.92709461],
           [ 0.74378508],
           [-3.2418159 ],
           [-1.32277347],
           [ 0.90118   ],
           [-0.19157262],
           [ 0.82855712],
           [ 0.24817994],
           [-1.43034777],
           [-0.21232344],
           [ 0.        ],
           [ 0.        ]])
    >>> sigma = 0.15
    >>> y = A.dot(alpha) + np.random.normal(scale=sigma, size=(A.shape[0], 1))
    >>> input_channel_params = {'tau': tau, 'theta_bar': 0, 'theta_tilde': 4,
    ... 'use_em': False}
    >>> config['input_channel_parameters'] = input_channel_params
    >>> output_channel_params = {'sigma_sq': sigma**2,
    ... 'noise_level_estimation': 'fixed'}
    >>> config['output_channel_parameters'] = output_channel_params
    >>> alpha_hat = run(y, A, A_asq)
    >>> alpha_hat[:k + 2]
    array([[ 1.93810961],
           [ 0.6955502 ],
           [-3.39759349],
           [-1.35533562],
           [ 1.10524227],
           [-0.00594848],
           [ 0.79274671],
           [ 0.04895264],
           [-1.08726071],
           [-0.00142911],
           [ 0.00022861],
           [-0.00004272]])
    >>> np.sum(np.abs(alpha - alpha_hat) > sigma * 3)
    0

    or recover the same vector returning a history comparing the pr. iteration
    solution to the true vector and printing the A_asq details

    >>> config['report_A_asq_setup'] = True
    >>> config['report_history'] = True
    >>> config['true_solution'] = alpha
    >>> alpha_hat, history = run(y, A, A_asq) # doctest: +NORMALIZE_WHITESPACE
    GAMP is using the A_asq: [[ 0.024 ..., 0.002]
     ...,
     [ 0. ..., 0.014]]
    The sum approximation method is: None
    >>> alpha_hat[:k + 2]
    array([[ 1.93810961],
           [ 0.6955502 ],
           [-3.39759349],
           [-1.35533562],
           [ 1.10524227],
           [-0.00594848],
           [ 0.79274671],
           [ 0.04895264],
           [-1.08726071],
           [-0.00142911],
           [ 0.00022861],
           [-0.00004272]])
    >>> np.array(history['MSE']).reshape(-1, 1)[1:11]
    array([[ 0.04562729],
           [ 0.01328304],
           [ 0.00112098],
           [ 0.00074968],
           [ 0.00080175],
           [ 0.00076615],
           [ 0.00077043]])

    or recover the same vector using sample variance AWGN noise level
    estimation

    >>> config['report_A_asq_setup'] = False
    >>> config['report_history'] = False
    >>> output_channel_params['noise_level_estimation'] = 'sample_variance'
    >>> config['output_channel_parameters'] = output_channel_params
    >>> alpha_hat = run(y, A, A_asq)
    >>> alpha_hat[:k + 2]
    array([[ 1.94820622],
           [ 0.72162206],
           [-3.39978431],
           [-1.35357001],
           [ 1.10701779],
           [-0.00834467],
           [ 0.79790879],
           [ 0.08441384],
           [-1.08946306],
           [-0.0015894 ],
           [ 0.00020561],
           [-0.00003623]])
    >>> np.sum(np.abs(alpha - alpha_hat) > sigma * 3)
    0

    or recover the same vector using median AWGN noise level estimation

    >>> output_channel_params['noise_level_estimation'] = 'median'
    >>> config['output_channel_parameters'] = output_channel_params
    >>> alpha_hat = run(y, A, A_asq)
    >>> alpha_hat[:k + 2]
    array([[ 1.93356483],
           [ 0.65232347],
           [-3.39440429],
           [-1.35437724],
           [ 1.10312573],
           [-0.0050555 ],
           [ 0.78743162],
           [ 0.03616397],
           [-1.08589927],
           [-0.00136802],
           [ 0.00024121],
           [-0.00004498]])
    >>> np.sum(np.abs(alpha - alpha_hat) > sigma * 3)
    0

    or recover the same vector learning the AWGN noise level using expectation
    maximization (EM)

    >>> output_channel_params['noise_level_estimation'] = 'em'
    >>> config['output_channel_parameters'] = output_channel_params
    >>> alpha_hat = run(y, A, A_asq)
    >>> alpha_hat[:k + 2]
    array([[ 1.94118089],
           [ 0.71553983],
           [-3.40076165],
           [-1.35662005],
           [ 1.1099417 ],
           [-0.00688125],
           [ 0.79442879],
           [ 0.06258856],
           [-1.08792606],
           [-0.00148811],
           [ 0.00022266],
           [-0.00003785]])
    >>> np.sum(np.abs(alpha - alpha_hat) > sigma * 3)
    0

    >>> np.set_printoptions(**np_printoptions)

    """

    @_decorate_validation
    def validate_input():
        _numeric('y', ('integer', 'floating', 'complex'), shape=(-1, 1))
        _numeric('A', ('integer', 'floating', 'complex'), shape=(
            y.shape[0],
            _conf['true_solution'].shape[0]
            if _conf['true_solution'] is not None else -1))

        if isinstance(A_asq, _MatrixBase):
            # It is not possible to validate the range of an implicit matrix
            # Thus allow all possible values
            range_ = '[-inf;inf]'
        else:
            range_ = '[0;inf)'
        _numeric('A_asq', ('integer', 'floating', 'complex'), range_=range_,
                 shape=A.shape, ignore_none=True)

    @_decorate_validation
    def validate_output():
        # complex128 is two float64 (real and imaginary part) each taking 8*8
        # bits. Thus, in total 2*8*8=128 bits. However, we only consider it to
        # be "64 bit precision" since that is what each part is.
        bits_pr_nbytes = 4 if np.iscomplexobj(convert(0)) else 8
        _numeric('alpha', ('integer', 'floating', 'complex'),
                 shape=(A.shape[1], 1),
                 precision=convert(0).nbytes * bits_pr_nbytes)
        _generic('history', 'mapping',
                 keys_in=('alpha_bar', 'alpha_tilde', 'MSE',
                          'input_channel_parameters',
                          'output_channel_parameters', 'stop_criterion',
                          'stop_criterion_value', 'stop_iteration',
                          'stop_reason'))

    validate_input()

    # Initialisation
    init = _get_gamp_initialisation(y, A, A_asq)
    AH = init['AH']
    A_asq = init['A_asq']
    AT_asq = init['AT_asq']
    o = init['o']
    q = init['q']
    s = init['s']
    r = init['r']
    m = init['m']
    n = init['n']
    alpha_bar = init['alpha_bar']
    alpha_tilde = init['alpha_tilde']
    alpha_breve = init['alpha_breve']
    A_dot_alpha_bar = init['A_dot_alpha_bar']
    z_bar = init['z_bar']
    damping = init['damping']
    sum_approximation_method = init['sum_approximation_method']
    A_frob_sq = init['A_frob_sq']
    output_channel = init['output_channel']
    output_channel_parameters = init['output_channel_parameters']
    input_channel = init['input_channel']
    input_channel_parameters = init['input_channel_parameters']
    stop_criterion = init['stop_criterion']
    stop_criterion_name = init['stop_criterion_name']
    iterations = init['iterations']
    tolerance = init['tolerance']
    convert = init['convert']
    report_history = init['report_history']
    history = init['history']
    true_solution = init['true_solution']

    # GAMP iterations
    for it in range(iterations):
        # Save previous state
        alpha_bar_prev = alpha_bar  # Used in stop criterion

        # GAMP state updates
        if sum_approximation_method == 'rangan':
            # Rangan's scalar variance sum approximation.

            # Factor side updates
            v = 1.0 / m * A_frob_sq * alpha_breve
            o = A_dot_alpha_bar - v * q
            z_bar, z_tilde = output_channel.compute(locals())
            q = (z_bar - o) / v
            u = 1.0 / m * np.sum((v - z_tilde) / v**2)

            # Variable side updates (including damping)
            s_full = 1 / (1.0 / n * A_frob_sq * u)
            s = (1 - damping) * s_full + damping * s
            r = alpha_bar + s * AH.dot(q)
            alpha_bar_full, alpha_tilde = input_channel.compute(locals())
            alpha_bar = (1 - damping) * alpha_bar_full + damping * alpha_bar
            alpha_breve = 1.0 / n * np.sum(alpha_tilde)

        else:
            # Either "full" GAMP or Krzakala's sum approximation.
            # For "full" GAMP, A_asq is the entrywise absolute value squared
            # matrix.
            # For Krzakala's sum approximation, A_asq is a
            # magni.utils.matrices.SumApproximationMatrix that implements the
            # scaled sum approximation to the full A_asq matrix.

            # Factor side updates
            v = A_asq.dot(alpha_tilde)
            o = A_dot_alpha_bar - v * q
            z_bar, z_tilde = output_channel.compute(locals())
            q = (z_bar - o) / v
            u = (v - z_tilde) / v**2

            # Variable side updates (including damping)
            s_full = 1 / (AT_asq.dot(u))
            s = (1 - damping) * s_full + damping * s
            r = alpha_bar + s * AH.dot(q)
            alpha_bar_full, alpha_tilde = input_channel.compute(locals())
            alpha_bar = (1 - damping) * alpha_bar_full + damping * alpha_bar

        # Stop criterion
        A_dot_alpha_bar = A.dot(alpha_bar)  # Used in residual stop criteria
        stop, stop_criterion_value = stop_criterion.compute(locals())

        # History reporting
        if report_history:
            history['alpha_bar'].append(alpha_bar)
            history['alpha_tilde'].append(alpha_tilde)
            history['input_channel_parameters'].append(
                copy.deepcopy(input_channel.__dict__))
            history['output_channel_parameters'].append(
                copy.deepcopy(output_channel.__dict__))
            history['stop_criterion_value'].append(stop_criterion_value)
            history['stop_iteration'] = it
            if true_solution is not None:
                history['MSE'].append(
                    1/n * np.linalg.norm(true_solution - alpha_bar)**2)

        if stop:
            history['stop_reason'] = stop_criterion_name.upper()
            break

    alpha = alpha_bar

    validate_output()

    if report_history:
        return alpha, history
    else:
        return alpha


def _get_gamp_initialisation(y, A, A_asq):
    """
    Return an initialisation of the GAMP algorithm.

    Parameters
    ----------
    y : ndarray
        The m x 1 measurement vector.
    A : ndarray or magni.utils.matrices.{Matrix, MatrixCollection}
        The m x n matrix which is the product of the measurement matrix and the
        dictionary matrix.
    A_asq : ndarray or magni.utils.matrices.{Matrix, MatrixCollection} or None
        The m x n matrix which is the entrywise absolute value squared product
        of the measurement matrix and the dictionary matrix (the default is
        None, which implies that a sum approximation is used).

    Returns
    -------
    init : dict
        The initialisation of the the GAMP algorithm.

    """

    init = dict()
    param = dict(_conf.items())

    # Configured setup
    init['convert'] = param['precision_float']
    init['damping'] = init['convert'](param['damping'])
    init['report_history'] = param['report_history']
    init['tolerance'] = param['tolerance']
    init['iterations'] = param['iterations']
    init['true_solution'] = param['true_solution']
    init['output_channel_parameters'] = param['output_channel_parameters']
    init['input_channel_parameters'] = param['input_channel_parameters']
    init['stop_criterion_name'] = param['stop_criterion'].__name__

    # Arguments based configuration
    init['AH'] = A.conj().T
    init['m'], init['n'] = A.shape

    init['A_frob_sq'] = None
    init['sum_approximation_method'] = None

    if A_asq is None:
        init['sum_approximation_method'], sum_approx_const = tuple(
            param['sum_approximation_constant'].items())[0]

        # Krzakala's sum approx
        init['A_asq'] = _SumApproximationMatrix(sum_approx_const)
        init['AT_asq'] = _SumApproximationMatrix(sum_approx_const)

        # Rangan's sum approx
        if init['sum_approximation_method'] == 'rangan':
            # _norm generally forms A explicitly to compute the Frobenius norm
            # which may be infeasible memory-wise
            init['A_frob_sq'] = _norm(A, ord='fro')**2

    else:
        # Full GAMP
        init['A_asq'] = A_asq
        init['AT_asq'] = A_asq.T

    # Channel and stop criterion configuration
    channel_init = copy.copy(init)
    channel_init['A'] = A
    channel_init['y'] = y
    init['input_channel'] = param['input_channel'](channel_init)
    init['output_channel'] = param['output_channel'](channel_init)
    init['stop_criterion'] = param['stop_criterion'](channel_init)

    # Configuration of remaining states
    if param['warm_start'] is not None:
        init['alpha_bar'] = init['convert'](param['warm_start'][0])
        init['alpha_tilde'] = init['convert'](param['warm_start'][1])
    else:
        init['alpha_bar'] = np.zeros((init['n'], 1), dtype=init['convert'])
        init['alpha_tilde'] = np.ones_like(init['alpha_bar'])

    init['alpha_breve'] = 1.0 / init['n'] * np.sum(init['alpha_tilde'])
    init['history'] = {'alpha_bar': [init['alpha_bar']],
                       'alpha_tilde': [init['alpha_tilde']],
                       'MSE': [np.nan],
                       'input_channel_parameters': [
                           init['input_channel_parameters']],
                       'output_channel_parameters': [
                           init['output_channel_parameters']],
                       'stop_criterion': init['stop_criterion_name'].upper(),
                       'stop_criterion_value': [np.nan],
                       'stop_iteration': 0,
                       'stop_reason': 'MAX_ITERATIONS'}
    init['A_dot_alpha_bar'] = A.dot(init['alpha_bar'])
    init['q'] = np.zeros_like(y)
    init['o'] = y.copy()
    init['s'] = np.ones_like(init['alpha_bar'])
    init['r'] = np.zeros_like(init['alpha_bar'])
    init['z_bar'] = y

    # Report on A_asq setup
    if param['report_A_asq_setup']:
        np_printoptions = np.get_printoptions()
        np.set_printoptions(precision=3, threshold=5, edgeitems=1)
        print('GAMP is using the A_asq: {}'.format(init['A_asq']))
        print('The sum approximation method is: {}'.format(
            init['sum_approximation_method']))
        np.set_printoptions(**np_printoptions)

    return init



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing the `validate_levels` function.

Routine listings
----------------
validate_levels(name, levels)
    Validate containers and mappings as well as contained objects.

"""

from __future__ import division

from magni.utils.validation import validate_generic as _generic
from magni.utils.validation import validate_numeric as _numeric
from magni.utils.validation._util import get_var as _get_var
from magni.utils.validation._util import report as _report


def validate_levels(name, levels):
    """
    Validate containers and mappings as well as contained objects

    The present function is meant to valdiate the 'levels' of a variable. That
    is, the value of the variable itself, the values of the second level (in
    case the value is a list, tuple, or dict), the values of the third level,
    and so on.

    Parameters
    ----------
    name : None
        The name of the variable to be validated.
    levels : list or tuple
        The list of levels.

    See Also
    --------
    magni.utils.validation.validate_generic : Validate non-numeric objects.
    magni.utils.validation.validate_numeric : Validate numeric objects.

    Notes
    -----
    `name` must refer to a variable in the parent scope of the function or
    method decorated by `magni.utils.validation.decorate_validation` which is
    closest to the top of the call stack. If `name` is a string then there must
    be a variable of that name in that scope. If `name` is a set-like object
    then there must be a variable having the first value in that set-like
    object as name. The remaining values are used as keys/indices on the
    variable to obtain the variable to be validated. For example, the `name`
    ('name', 0, 'key') refers to the variable "name[0]['key']".

    `levels` is a list containing the levels. The value of the variable is
    validated against the first level. In case the value is a list, tuple, or
    dict, the values contained in this are validated against the second level
    and so on. Each level is itself a list with the first value being either
    'generic' or 'numeric' followed by the arguments that should be passed to
    the respective function (with the exception of `name` which is
    automatically prepended by the present function).

    Examples
    --------
    Every public function and method of the present package (with the exception
    of the functions of this subpackage itself) validates every argument and
    keyword argument using the functionality of this subpackage. Thus, for
    examples of how to use the present function, browse through the code.

    """

    if name is None:
        return ('levels', levels)

    if isinstance(name, (list, tuple)):
        name = list(name)
    else:
        name = [name]

    if not isinstance(levels, (list, tuple)):
        _report(TypeError, 'must be in {!r}.', (list, tuple),
                var_name='levels', var_value=levels, expr='type({})',
                prepend='Invalid validation call: ')

    if len(levels) == 0:
        _report(ValueError, 'must be > 0.', var_name='levels',
                var_value=levels, expr='len({})',
                prepend='Invalid validation call: ')

    _validate_level(name, _get_var(name), levels)


def _validate_level(name, var, levels, index=0):
    """
    Validate a level.

    Parameters
    ----------
    name : None
        The name of the variable.
    var : None
        The value of the variable.
    levels : set-like
        The levels.
    index : int
        The index of the current level. (the default is 0)

    """

    if not isinstance(levels[index], (list, tuple)):
        _report(TypeError, 'must be in {!r}.', (list, tuple),
                var_name=('levels', index), var_value=levels[index],
                expr='type({})', prepend='Invalid validation call: ')

    if len(levels[index]) == 0:
        _report(ValueError, 'must be > 0.', var_name=('levels', index),
                var_value=levels[index], expr='len({})',
                prepend='Invalid validation call: ')

    if levels[index][0] not in ('generic', 'numeric'):
        _report(ValueError, 'must be in {!r}.', ('generic', 'numeric'),
                var_name=('levels', index, 0), var_value=levels[index][0],
                prepend='Invalid validation call: ')

    func = {'generic': _generic, 'numeric': _numeric}[levels[index][0]]
    args = (name,) + tuple(levels[index][1:]) + (var,)
    func(*args)

    if index + 1 < len(levels):
        if isinstance(var, (list, tuple)):
            for i, value in enumerate(var):
                _validate_level(name + [i], value, levels, index + 1)
        elif isinstance(var, dict):
            for key, value in var.items():
                _validate_level(name + [key], value, levels, index + 1)



"""
..
    Copyright (c) 2015-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing unittests for `magni.cs.reconstruction`.

**Testing Strategy**
The usage of FastOps is tested along with reconstructions in various points in
the phase space.

**Phase Space Tests Overview**
A set of :math:`(\delta, \rho)` points in the phase space is selected. For each
algorithm, the reconstruction capabilities in each point has been determined
for a given problem suite. The tests are based on a comparison with these
reference results. Specifically, a comparison based on `np.allclose` is done.
Also some border cases (extremes) like :math:`k = 0` are tested.


Points where it is likely to have positive results
+-----+------+------+------+------+------+------+------+
|  no.|   1  |   2  |   3  |   4  |   5  |   6  |   7  |
+-----+------+------+------+------+------+------+------+
|delta| 0.08 | 0.24 | 0.38 | 0.62 | 0.78 | 0.84 | 0.96 |
+-----+------+------+------+------+------+------+------+
|  rho| 0.05 | 0.01 | 0.12 | 0.38 | 0.22 | 0.08 | 0.91 |
+-----+------+------+------+------+------+------+------+


Points where it is unlikely to have positive results
+-----+------+------+------+
|  no.|   A  |   B  |   C  |
+-----+------+------+------+
|delta| 0.06 | 0.19 | 0.29 |
+-----+------+------+------+
|  rho| 0.92 | 0.84 | 0.94 |
+-----+------+------+------+


**Functions Tested**

See the docstrings of the below listed classes.

Routine listings
----------------
ComparisonGAMPTests(unittest.TestCase)
    Comparison of magni.cs.reconstruction.gamp to a reference implementation.
FastOpsTests(unittest.TestCase)
    Tests of FastOp input, i.e. function based measurements and FFT dictionary.
FeatureTest(object)
    Reconstruction algorithm feature test base class.
FeaturePrecisionFloatTest(FeatureTest, unittest.TestCase)
    Test of the precision float feature in reconstruction algorithms.
FeatureReportHistoryTest(FeatureTest, unittest.TestCase)
    Test the report history feature in reconstruction algorithms.
FeatureStopCriterionTest(FeatureTest, unittest.TestCase)
    Test of the stop criterion feature in reconstruction algorithms.
FeatureWarmStartTest(FeatureTest, unittest.TestCase)
    Test of the warm_start feature in reconstruction algorithms.
PhaseSpaceExtremesTest(unittest.TestCase):
    Tests of border case (extreme) phase space values.
PhaseSpaceTest(object)
    Phase space test base class.
PhaseSpaceTest1(PhaseSpaceTest, unittest.TestCase)
    Test of reconstruction capabilities at Phase Space point (0.08, 0.05)
PhaseSpaceTest2(PhaseSpaceTest, unittest.TestCase)
    Test of reconstruction capabilities at Phase Space point (0.24, 0.01)
PhaseSpaceTest3(PhaseSpaceTest, unittest.TestCase)
    Test of reconstruction capabilities at Phase Space point (0.38, 0.12)
PhaseSpaceTest4(PhaseSpaceTest, unittest.TestCase)
    Test of reconstruction capabilities at Phase Space point (0.62, 0.38)
PhaseSpaceTest5(PhaseSpaceTest, unittest.TestCase)
    Test of reconstruction capabilities at Phase Space point (0.78, 0.22)
PhaseSpaceTest6(PhaseSpaceTest, unittest.TestCase)
    Test of reconstruction capabilities at Phase Space point (0.84, 0.08)
PhaseSpaceTest7(PhaseSpaceTest, unittest.TestCase)
    Test of reconstruction capabilities at Phase Space point (0.96, 0.91)
PhaseSpaceTestA(PhaseSpaceTest, unittest.TestCase)
    Test of reconstruction capabilities at Phase Space point (0.06, 0.92)
PhaseSpaceTestB(PhaseSpaceTest, unittest.TestCase)
    Test of reconstruction capabilities at Phase Space point (0.19, 0.84)
PhaseSpaceTestC(PhaseSpaceTest, unittest.TestCase)
    Test of reconstruction capabilities at Phase Space point (0.29, 0.94)
TestUSERademacher(unittest.TestCase)
    Test of the use_rademacher test fixture function.
use_rademacher(n, m, k, seed)
    Prepare an instance of the USE/Rademacher problem suite

"""

from __future__ import division
import os
import unittest
import warnings

import numpy as np

import magni
from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate_numeric as _numeric


class ComparisonGAMPTests(unittest.TestCase):
    """
    Comparison of magni.cs.reconstruction.gamp to a reference implementation.

    **Reference implementation**
    "run_amp" from
    https://github.com/eric-tramel/SwAMP-Demo/blob/master/python/amp.py
    commit b32755caa8d6b59929174e2a06cc685bae5849b6

    """

    def setUp(self):
        self.ns = [1024, 2048, 2048, 2000, 1000]
        self.ms = [770, 780, 1024, 800, 500]
        self.ks = [440, 440, 1024, 126, 88]
        self.sigma_sqs = [0.0, 0.0, 0.0, 1e-3, 1e-2]
        self.sigma_sqs_init = [1e-6, 1e-6, 1e-6, 1, 1]
        self.theta_bars = [0.0, 0.0, 0.0, 0.2, 0.0]
        self.theta_tildes = [1.0, 1.0, 1.0, 0.3, 1.0]
        self.taus = [float(k) / n for (k, n) in zip(self.ks, self.ns)]

        self.tolerance = 1e-16
        self.iterations = 500

        np.random.seed(6021)
        self.seeds = np.random.randint(1000, 80000, size=len(self.ns))

        file_path = os.path.dirname(os.path.abspath(__file__)) + os.sep
        fixed_sigma_sq_sol_file = np.load(
            file_path + 'gamp_fixed_sigma_sq_sols.npz')
        em_sigma_sq_sol_file = np.load(
            file_path + 'gamp_em_sigma_sq_sols.npz')
        self.comparison_solutions_fixed_sigma_sq = [
            fixed_sigma_sq_sol_file[arr] for arr in sorted(
                fixed_sigma_sq_sol_file.files)]
        self.comparison_solutions_em_sigma_sq = [
            em_sigma_sq_sol_file[arr] for arr in sorted(
                em_sigma_sq_sol_file.files)]

    def tearDown(self):
        magni.cs.reconstruction.gamp.config.reset()

    def testFixedGAMPComparison(self):
        for ix in range(len(self.ns)):
            # Setup GAMP solver
            input_channel_params = {'tau': self.taus[ix],
                                    'theta_bar': self.theta_bars[ix],
                                    'theta_tilde': self.theta_tildes[ix],
                                    'use_em': False}
            output_channel_params = {
                'sigma_sq': self.sigma_sqs_init[ix],
                'noise_level_estimation': 'fixed'}
            gamp_config = {'tolerance': self.tolerance,
                           'iterations': self.iterations,
                           'input_channel_parameters': input_channel_params,
                           'output_channel_parameters': output_channel_params}
            magni.cs.reconstruction.gamp.config.update(gamp_config)

            # Generate problem instance
            z, A, alpha = use_gaussian(
                self.ns[ix], self.ms[ix], self.ks[ix], self.seeds[ix])
            A_asq = np.abs(A)**2
            if self.sigma_sqs[ix] > 0:
                y = z + np.random.normal(
                    size=z.shape, loc=0.0, scale=np.sqrt(self.sigma_sqs[ix]))
            else:
                y = z

            # Run solver
            alpha_hat = magni.cs.reconstruction.gamp.run(y, A, A_asq)

            # Compare result
            self.assertTrue(
                np.allclose(
                    self.comparison_solutions_fixed_sigma_sq[ix],
                    alpha_hat.flatten()))

    def testEMGAMPComparison(self):
        for ix in range(len(self.ns)):
            # Setup GAMP solver
            input_channel_params = {'tau': self.taus[ix],
                                    'theta_bar': self.theta_bars[ix],
                                    'theta_tilde': self.theta_tildes[ix],
                                    'use_em': False}
            output_channel_params = {
                'sigma_sq': self.sigma_sqs_init[ix],
                'noise_level_estimation': 'em'}
            gamp_config = {'tolerance': self.tolerance,
                           'iterations': self.iterations,
                           'input_channel_parameters': input_channel_params,
                           'output_channel_parameters': output_channel_params}
            magni.cs.reconstruction.gamp.config.update(gamp_config)

            # Generate problem instance
            z, A, alpha = use_gaussian(
                self.ns[ix], self.ms[ix], self.ks[ix], self.seeds[ix])
            A_asq = np.abs(A)**2
            if self.sigma_sqs[ix] > 0:
                y = z + np.random.normal(
                    size=z.shape, loc=0.0, scale=np.sqrt(self.sigma_sqs[ix]))
            else:
                y = z

            # Run solver
            alpha_hat = magni.cs.reconstruction.gamp.run(y, A, A_asq)

            # Compare result
            # The difference in EM learning for AMP vs Symmetric GAMP with AWGN
            # output channel makes it difficult to compare the results.
            # Thus, we only compare the non-zeros up to atol=1e-7.
            self.assertTrue(
                np.allclose(
                    self.comparison_solutions_em_sigma_sq[ix][:self.ks[ix]],
                    alpha_hat.flatten()[:self.ks[ix]],
                    atol=1e-7))


class FastOpsTests(unittest.TestCase):
    """
    Tests of FastOp input, i.e. function based measurements and FFT dictionary.

    The following tests are implemented:

    - *test_AMP_with_DCT_FFT_vs_Separable_2D*
    - *test_GAMP_with_DCT_FFT_vs_Separable_2D_rangan_sum_approx*
    - *test_GAMP_with_DCT_FFT_vs_Separable_2D_krzakala_sum_approx*
    - *test_GAMP_with_DCT_Separable_full_transform_and_precision*
    - *test_IT_with_DCT*
    - *test_IT_with_DFT*
    - *test_IT_with_DCT_and_precision*

    """

    def setUp(self):
        h = 25
        w = 25
        n = h * w
        k = 15

        self.problem_dim = (h, w)

        # Spiral scan pattern
        scan_length = 0.30 * 2 * h * w
        num_points = 10 * int(scan_length)
        img_coords = magni.imaging.measurements.spiral_sample_image(
            h, w, scan_length, num_points, rect_area=True)
        self.Phi = magni.imaging.measurements.construct_measurement_matrix(
            img_coords, h, w)

        np.random.seed(6021)
        self.alpha_real = np.zeros((n, 1))
        self.alpha_real[:k, 0] = np.random.normal(size=k, loc=2, scale=2.0)
        self.alpha_complex = np.zeros((n, 1), dtype=np.complex128)
        self.alpha_complex[:k, 0] = (np.random.randn(k) +
                                     1j * np.random.randn(k))

        self.noise = np.random.normal(size=(self.Phi.shape[0], 1), scale=0.01)

        magni.cs.reconstruction.it.config.update(
            {'iterations': 200, 'threshold': 'fixed', 'threshold_fixed': k})
        magni.cs.reconstruction.gamp.config.update(
            {'iterations': 500,
             'tolerance': 1e-6,
             'input_channel_parameters': {'tau': k/n,
                                          'theta_bar': 2.0,
                                          'theta_tilde': 4.0,
                                          'use_em': False},
             'output_channel_parameters': {'sigma_sq': 1.0,
                                           'noise_level_estimation': 'median'}}
        )

    def tearDown(self):
        magni.cs.reconstruction.it.config.reset()
        magni.cs.reconstruction.amp.config.reset()
        magni.cs.reconstruction.gamp.config.reset()

    def test_AMP_with_DCT_FFT_vs_Separable_2D(self):
        Psi_fft = magni.imaging.dictionaries.get_DCT(self.problem_dim)
        A_fft = magni.utils.matrices.MatrixCollection((self.Phi, Psi_fft))
        y_fft = A_fft.dot(self.alpha_real) + self.noise

        iDCT_mtx = magni.imaging.dictionaries.get_DCT_transform_matrix(
            self.problem_dim[0]).T
        Psi_sep = magni.utils.matrices.Separable2DTransform(iDCT_mtx, iDCT_mtx)
        A_sep = magni.utils.matrices.MatrixCollection((self.Phi, Psi_sep))
        y_sep = A_sep.dot(self.alpha_real) + self.noise

        self.assertTrue(np.allclose(y_fft, y_sep))
        alpha_hat_fft = self._amp_run(y_fft, A_fft, self.alpha_real,
                                      success=True)
        alpha_hat_sep = self._amp_run(y_sep, A_sep, self.alpha_real,
                                      success=True)
        self.assertTrue(np.allclose(alpha_hat_fft, alpha_hat_sep))

    def test_GAMP_with_DCT_FFT_vs_Separable_2D_rangan_sum_approx(self):
        Psi_fft = magni.imaging.dictionaries.get_DCT(self.problem_dim)
        A_fft = magni.utils.matrices.MatrixCollection((self.Phi, Psi_fft))
        y_fft = A_fft.dot(self.alpha_real) + self.noise

        iDCT_mtx = magni.imaging.dictionaries.get_DCT_transform_matrix(
            self.problem_dim[0]).T
        Psi_sep = magni.utils.matrices.Separable2DTransform(iDCT_mtx, iDCT_mtx)
        A_sep = magni.utils.matrices.MatrixCollection((self.Phi, Psi_sep))
        y_sep = A_sep.dot(self.alpha_real) + self.noise

        self.assertEqual(
            magni.cs.reconstruction.gamp.config['sum_approximation_constant'],
            {'rangan': 1.0})
        self.assertTrue(np.allclose(y_fft, y_sep))
        alpha_hat_fft = self._gamp_run(
            y_fft, A_fft, None, self.alpha_real, success=True)
        alpha_hat_sep = self._gamp_run(
            y_sep, A_sep, None, self.alpha_real, success=True)
        self.assertTrue(np.allclose(alpha_hat_fft, alpha_hat_sep))

    def test_GAMP_with_DCT_FFT_vs_Separable_2D_krzakala_sum_approx(self):
        magni.cs.reconstruction.gamp.config['sum_approximation_constant'] = {
            'krzakala': 1.0 / (self.problem_dim[0] * self.problem_dim[1])}
        Psi_fft = magni.imaging.dictionaries.get_DCT(self.problem_dim)
        A_fft = magni.utils.matrices.MatrixCollection((self.Phi, Psi_fft))
        y_fft = A_fft.dot(self.alpha_real) + self.noise

        iDCT_mtx = magni.imaging.dictionaries.get_DCT_transform_matrix(
            self.problem_dim[0]).T
        Psi_sep = magni.utils.matrices.Separable2DTransform(iDCT_mtx, iDCT_mtx)
        A_sep = magni.utils.matrices.MatrixCollection((self.Phi, Psi_sep))
        y_sep = A_sep.dot(self.alpha_real) + self.noise

        self.assertEqual(
            magni.cs.reconstruction.gamp.config['sum_approximation_constant'],
            {'krzakala': 1.0 / (self.problem_dim[0] * self.problem_dim[1])})
        self.assertTrue(np.allclose(y_fft, y_sep))
        alpha_hat_fft = self._gamp_run(
            y_fft, A_fft, None, self.alpha_real, success=True)
        alpha_hat_sep = self._gamp_run(
            y_sep, A_sep, None, self.alpha_real, success=True)
        self.assertTrue(np.allclose(alpha_hat_fft, alpha_hat_sep))

    def test_GAMP_with_DCT_Separable_full_transform_and_precision(self):
        # Float 32
        magni.cs.reconstruction.gamp.config['precision_float'] = np.float32
        self.assertEqual(
            magni.cs.reconstruction.gamp.config['precision_float'], np.float32)
        iDCT_mtx = np.float32(
            magni.imaging.dictionaries.get_DCT_transform_matrix(
                self.problem_dim[0]).T)
        iDCT_mtx_asq = np.abs(iDCT_mtx) ** 2
        Psi = magni.utils.matrices.Separable2DTransform(iDCT_mtx, iDCT_mtx)
        Psi_asq = magni.utils.matrices.Separable2DTransform(iDCT_mtx_asq,
                                                            iDCT_mtx_asq)
        A = magni.utils.matrices.MatrixCollection((self.Phi, Psi))
        A_asq = magni.utils.matrices.MatrixCollection((self.Phi, Psi_asq))
        y = A.dot(np.float32(self.alpha_real)) + np.float32(self.noise)

        self.assertEqual(y.dtype, np.float32)
        self.assertEqual(A.T.dot(y).dtype, np.float32)
        self.assertEqual(A_asq.T.dot(y).dtype, np.float32)
        self.assertEqual(A_asq.A.dtype, (A.A**2).dtype)
        self.assertTrue(np.allclose(A_asq.A, A.A**2))
        alpha_hat = self._gamp_run(y, A, A_asq, self.alpha_real)
        self.assertEqual(alpha_hat.dtype, np.float32)

        # Float64
        magni.cs.reconstruction.gamp.config['precision_float'] = np.float64
        self.assertEqual(
            magni.cs.reconstruction.gamp.config['precision_float'], np.float64)
        iDCT_mtx = np.float64(
            magni.imaging.dictionaries.get_DCT_transform_matrix(
                self.problem_dim[0]).T)
        iDCT_mtx_asq = np.abs(iDCT_mtx) ** 2
        Psi = magni.utils.matrices.Separable2DTransform(iDCT_mtx, iDCT_mtx)
        Psi_asq = magni.utils.matrices.Separable2DTransform(iDCT_mtx_asq,
                                                            iDCT_mtx_asq)
        A = magni.utils.matrices.MatrixCollection((self.Phi, Psi))
        A_asq = magni.utils.matrices.MatrixCollection((self.Phi, Psi_asq))
        y = A.dot(np.float64(self.alpha_real)) + np.float64(self.noise)

        self.assertEqual(y.dtype, np.float64)
        self.assertEqual(A.T.dot(y).dtype, np.float64)
        self.assertEqual(A_asq.T.dot(y).dtype, np.float64)
        self.assertEqual(A_asq.A.dtype, (A.A**2).dtype)
        self.assertTrue(np.allclose(A_asq.A, A.A**2))
        alpha_hat = self._gamp_run(y, A, A_asq, self.alpha_real)
        self.assertEqual(alpha_hat.dtype, np.float64)

        # Float128
        if not hasattr(np, 'float128'):
            return

        magni.cs.reconstruction.gamp.config['precision_float'] = np.float128
        self.assertEqual(
            magni.cs.reconstruction.gamp.config['precision_float'],
            np.float128)
        iDCT_mtx = np.float128(
            magni.imaging.dictionaries.get_DCT_transform_matrix(
                self.problem_dim[0]).T)
        iDCT_mtx_asq = np.abs(iDCT_mtx) ** 2
        Psi = magni.utils.matrices.Separable2DTransform(iDCT_mtx, iDCT_mtx)
        Psi_asq = magni.utils.matrices.Separable2DTransform(iDCT_mtx_asq,
                                                            iDCT_mtx_asq)
        A = magni.utils.matrices.MatrixCollection((self.Phi, Psi))
        A_asq = magni.utils.matrices.MatrixCollection((self.Phi, Psi_asq))
        y = A.dot(np.float128(self.alpha_real)) + np.float64(self.noise)

        self.assertEqual(y.dtype, np.float128)
        self.assertEqual(A.T.dot(y).dtype, np.float128)
        self.assertEqual(A_asq.T.dot(y).dtype, np.float128)
        self.assertEqual(A_asq.A.dtype, (A.A**2).dtype)
        self.assertTrue(np.allclose(A_asq.A, A.A**2))
        alpha_hat = self._gamp_run(y, A, A_asq, self.alpha_real)
        self.assertEqual(alpha_hat.dtype, np.float128)

    def test_IT_with_DCT_and_precision(self):
        Psi = magni.imaging.dictionaries.get_DCT(self.problem_dim)
        A = magni.utils.matrices.MatrixCollection((self.Phi, Psi))

        # Float 32
        magni.cs.reconstruction.it.config.update(
            {'precision_float': np.float32})
        self.assertEqual(
            magni.cs.reconstruction.it.config['precision_float'], np.float32)
        y = A.dot(np.float32(self.alpha_real))
        self.assertEqual(y.dtype, np.float32)
        self.assertEqual(A.T.dot(y).dtype, np.float32)
        alpha_hat = self._iht_run(y, A, self.alpha_real)
        self.assertEqual(alpha_hat.dtype, np.float32)
        alpha_hat = self._ist_run(y, A, self.alpha_real)
        self.assertEqual(alpha_hat.dtype, np.float32)

        # Float 64
        magni.cs.reconstruction.it.config.update(
            {'precision_float': np.float64})
        self.assertEqual(
            magni.cs.reconstruction.it.config['precision_float'], np.float64)
        y = A.dot(np.float64(self.alpha_real))
        self.assertEqual(y.dtype, np.float64)
        self.assertEqual(A.T.dot(y).dtype, np.float64)
        alpha_hat = self._iht_run(y, A, self.alpha_real)
        self.assertEqual(alpha_hat.dtype, np.float64)
        alpha_hat = self._ist_run(y, A, self.alpha_real)
        self.assertEqual(alpha_hat.dtype, np.float64)

        # Scipy DCT does not support float128

    def test_IT_with_DCT(self):
        Psi = magni.imaging.dictionaries.get_DCT(self.problem_dim)
        A = magni.utils.matrices.MatrixCollection((self.Phi, Psi))
        y = A.dot(self.alpha_real)

        self._iht_run(y, A, self.alpha_real)
        self._ist_run(y, A, self.alpha_real)

    def test_IT_with_DFT(self):
        Psi = magni.imaging.dictionaries.get_DFT(self.problem_dim)
        A = magni.utils.matrices.MatrixCollection((self.Phi, Psi))
        y_real = A.dot(self.alpha_real)
        y_complex = A.dot(self.alpha_complex)

        self._iht_run(y_real, A, self.alpha_real)
        self._ist_run(y_real, A, self.alpha_real)
        self._iht_run(y_complex, A, self.alpha_complex)
        self._ist_run(y_complex, A, self.alpha_complex)

    def _amp_run(self, y, A, a, success=True):
        threshold_params = {
            'theta': magni.cs.reconstruction.amp.util.theta_mm(
                float(A.shape[0]) / A.shape[1]), 'tau_hat_sq': 1.0,
            'threshold_level_update_method': 'residual'}
        magni.cs.reconstruction.amp.config['threshold_parameters'].update(
            threshold_params)
        a_hat = magni.cs.reconstruction.amp.run(y, A)

        if success:
            self.assertTrue(np.allclose(a_hat, a, atol=1e-1))
        else:
            self.assertFalse(np.allclose(a_hat, a, atol=1e-1))

        return a_hat

    def _gamp_run(self, y, F, F_sq, a, success=True):
        a_hat = magni.cs.reconstruction.gamp.run(y, F, F_sq)
        if success:
            self.assertTrue(np.allclose(a_hat, a, atol=1e-1))
        else:
            self.assertFalse(np.allclose(a_hat, a, atol=1e-1))

        return a_hat

    def _iht_run(self, y, A, alpha, success=True):
        iht_config = {'threshold_operator': 'hard'}
        magni.cs.reconstruction.it.config.update(iht_config)
        self.assertEqual(
            magni.cs.reconstruction.it.config['threshold_operator'], 'hard')

        alpha_hat = magni.cs.reconstruction.it.run(y, A)
        if success:
            self.assertTrue(np.allclose(alpha_hat, alpha, atol=1e-2))
        else:
            self.assertFalse(np.allclose(alpha_hat, alpha, atol=1e-2))

        return alpha_hat

    def _ist_run(self, y, A, alpha, success=True):
        ist_config = {'threshold_operator': 'soft'}
        magni.cs.reconstruction.it.config.update(ist_config)
        self.assertEqual(
            magni.cs.reconstruction.it.config['threshold_operator'], 'soft')

        alpha_hat = magni.cs.reconstruction.it.run(y, A)

        if success:
            self.assertTrue(np.allclose(alpha_hat, alpha, atol=1e-2))
        else:
            self.assertFalse(np.allclose(alpha_hat, alpha, atol=1e-2))

        return alpha_hat


class FeatureTest(object):
    """
    Reconstruction algorithm feature test base class.

    This class defines a reconstruction problem which may be used as the base
    for testing features of reconstruction algorithms such as warm start or
    different stop criteria.

    See the individual feature test classes for further information.

    """

    def setUp(self):
        seed = 6021
        n = 500
        delta = 0.68
        rho = 0.17
        m = int(delta * n)

        self.k = int(rho * m)
        self.tau = delta * rho

        self.y, self.A, self.alpha = use_rademacher(n, m, self.k, seed=seed)
        self.oracle_support = self.alpha != 0

        self.z, self.F, self.a = use_gaussian(n, m, self.k, seed=seed)
        self.F_sq = self.F**2

        magni.cs.reconstruction.it.config.update(iterations=200)
        magni.cs.reconstruction.gamp.config.update(iterations=200)

    def tearDown(self):
        magni.cs.reconstruction.it.config.reset()
        magni.cs.reconstruction.amp.config.reset()
        magni.cs.reconstruction.gamp.config.reset()

    def _amp_run(self, y, A, a, success=True):
        threshold_params = {
            'theta': magni.cs.reconstruction.amp.util.theta_mm(
                float(A.shape[0]) / A.shape[1]), 'tau_hat_sq': 1.0,
            'threshold_level_update_method': 'residual'}
        magni.cs.reconstruction.amp.config['threshold_parameters'].update(
            threshold_params)
        a_hat = magni.cs.reconstruction.amp.run(y, A)

        if success:
            self.assertTrue(np.allclose(a_hat, a, atol=1e-2))
        else:
            self.assertFalse(np.allclose(a_hat, a, atol=1e-2))

        return a_hat

    def _amp_history_run(self, y, A, a, success=True):
        threshold_params = {
            'theta': magni.cs.reconstruction.amp.util.theta_mm(
                float(A.shape[0]) / A.shape[1]), 'tau_hat_sq': 1.0,
            'threshold_level_update_method': 'residual'}
        magni.cs.reconstruction.amp.config['threshold_parameters'].update(
            threshold_params)
        a_hat, history = magni.cs.reconstruction.amp.run(y, A)

        if success:
            self.assertTrue(np.allclose(a_hat, a, atol=1e-2))
        else:
            self.assertFalse(np.allclose(a_hat, a, atol=1e-2))

        return history

    def _gamp_run(self, z, F, F_sq, a, success=True):
        a_hat = magni.cs.reconstruction.gamp.run(z, F, F_sq)

        if success:
            self.assertTrue(np.allclose(a_hat, a, atol=1e-2))
        else:
            self.assertFalse(np.allclose(a_hat, a, atol=1e-2))

        return a_hat

    def _gamp_history_run(self, z, F, F_sq, a, success=True):
        a_hat, history = magni.cs.reconstruction.gamp.run(z, F, F_sq)

        if success:
            self.assertTrue(np.allclose(a_hat, a, atol=1e-2))
        else:
            self.assertFalse(np.allclose(a_hat, a, atol=1e-2))

        return history

    def _iht_run(self, y, A, alpha, success=True):
        iht_config = {'threshold_operator': 'hard'}
        magni.cs.reconstruction.it.config.update(iht_config)
        self.assertEqual(
            magni.cs.reconstruction.it.config['threshold_operator'], 'hard')

        alpha_hat = magni.cs.reconstruction.it.run(y, A)

        if success:
            self.assertTrue(np.allclose(alpha_hat, alpha, atol=1e-2))
        else:
            self.assertFalse(np.allclose(alpha_hat, alpha, atol=1e-2))

        return alpha_hat

    def _ist_run(self, y, A, alpha, success=True):
        ist_config = {'threshold_operator': 'soft'}
        magni.cs.reconstruction.it.config.update(ist_config)
        self.assertEqual(
            magni.cs.reconstruction.it.config['threshold_operator'], 'soft')

        alpha_hat = magni.cs.reconstruction.it.run(y, A)

        if success:
            self.assertTrue(np.allclose(alpha_hat, alpha, atol=1e-2))
        else:
            self.assertFalse(np.allclose(alpha_hat, alpha, atol=1e-2))

        return alpha_hat

    def _ist_history_run(self, y, A, alpha, success=True):
        ist_config = {'threshold_operator': 'soft'}
        magni.cs.reconstruction.it.config.update(ist_config)
        self.assertEqual(
            magni.cs.reconstruction.it.config['threshold_operator'], 'soft')

        alpha_hat, history = magni.cs.reconstruction.it.run(y, A)

        if success:
            self.assertTrue(np.allclose(alpha_hat, alpha, atol=1e-2))
        else:
            self.assertFalse(np.allclose(alpha_hat, alpha, atol=1e-2))

        return history


class FeaturePrecisionFloatTest(FeatureTest, unittest.TestCase):
    """
    Test of the precision float feature in reconstruction algorithms.

    The following tests are implemented:

    - *test_float32_AMP*
    - *test_float32_GAMP*
    - *test_float64_AMP*
    - *test_float64_GAMP*
    - *test_float128_AMP*
    - *test_float128_GAMP*
    - *test_float32_IST*
    - *test_float64_IST*
    - *test_float128_IST*

    """

    def test_float32_AMP(self, success=True):
        magni.cs.reconstruction.amp.config['precision_float'] = np.float32
        self.y = np.float32(self.y)
        self.A = np.float32(self.A)
        self.assertEqual(
            magni.cs.reconstruction.amp.config['precision_float'], np.float32)

        a_hat = self._amp_run(self.y, self.A, self.alpha, success=success)
        self.assertEqual(a_hat.dtype, np.float32)

    def test_float32_GAMP(self, success=True):
        input_channel_params = {'tau': self.tau, 'theta_bar': 0,
                                'theta_tilde': 1, 'use_em': False}
        output_channel_params = {'sigma_sq': 1,
                                 'noise_level_estimation': 'sample_variance'}
        sc = magni.cs.reconstruction.gamp.stop_criterion
        magni.cs.reconstruction.gamp.config.update(
            {'input_channel_parameters': input_channel_params,
             'output_channel_parameters': output_channel_params,
             'stop_criterion': sc.Residual,
             'precision_float': np.float32})
        self.z = np.float32(self.z)
        self.F = np.float32(self.F)
        self.F_sq = np.float32(self.F_sq)
        self.assertEqual(
            magni.cs.reconstruction.gamp.config['precision_float'], np.float32)

        a_hat = self._gamp_run(
            self.z, self.F, self.F_sq, self.a, success=success)
        self.assertEqual(a_hat.dtype, np.float32)

    def test_float32_GAMP_EM(self, success=True):
        input_channel_params = {'tau': self.tau, 'theta_bar': 0,
                                'theta_tilde': 1, 'use_em': True}
        output_channel_params = {'sigma_sq': 1,
                                 'noise_level_estimation': 'em'}
        sc = magni.cs.reconstruction.gamp.stop_criterion
        magni.cs.reconstruction.gamp.config.update(
            {'input_channel_parameters': input_channel_params,
             'output_channel_parameters': output_channel_params,
             'stop_criterion': sc.Residual,
             'precision_float': np.float32})
        self.z = np.float32(self.z)
        self.F = np.float32(self.F)
        self.F_sq = np.float32(self.F_sq)
        self.assertEqual(
            magni.cs.reconstruction.gamp.config['precision_float'], np.float32)

        a_hat = self._gamp_run(
            self.z, self.F, self.F_sq, self.a, success=success)
        self.assertEqual(a_hat.dtype, np.float32)

    def test_float64_AMP(self, success=True):
        magni.cs.reconstruction.amp.config['precision_float'] = np.float64
        self.y = np.float64(self.y)
        self.A = np.float64(self.A)
        self.assertEqual(
            magni.cs.reconstruction.amp.config['precision_float'], np.float64)

        a_hat = self._amp_run(self.y, self.A, self.alpha, success=success)
        self.assertEqual(a_hat.dtype, np.float64)

    def test_float64_GAMP(self, success=True):
        input_channel_params = {'tau': self.tau, 'theta_bar': 0,
                                'theta_tilde': 1, 'use_em': False}
        output_channel_params = {'sigma_sq': 1,
                                 'noise_level_estimation': 'sample_variance'}
        sc = magni.cs.reconstruction.gamp.stop_criterion
        magni.cs.reconstruction.gamp.config.update(
            {'input_channel_parameters': input_channel_params,
             'output_channel_parameters': output_channel_params,
             'stop_criterion': sc.Residual,
             'precision_float': np.float64})
        self.z = np.float64(self.z)
        self.F = np.float64(self.F)
        self.F_sq = np.float64(self.F_sq)
        self.assertEqual(
            magni.cs.reconstruction.gamp.config['precision_float'], np.float64)

        a_hat = self._gamp_run(
            self.z, self.F, self.F_sq, self.a, success=success)
        self.assertEqual(a_hat.dtype, np.float64)

    def test_float64_GAMP_EM(self, success=True):
        input_channel_params = {'tau': self.tau, 'theta_bar': 0,
                                'theta_tilde': 1, 'use_em': True}
        output_channel_params = {'sigma_sq': 1,
                                 'noise_level_estimation': 'em'}
        sc = magni.cs.reconstruction.gamp.stop_criterion
        magni.cs.reconstruction.gamp.config.update(
            {'input_channel_parameters': input_channel_params,
             'output_channel_parameters': output_channel_params,
             'stop_criterion': sc.Residual,
             'precision_float': np.float64})
        self.z = np.float64(self.z)
        self.F = np.float64(self.F)
        self.F_sq = np.float64(self.F_sq)
        self.assertEqual(
            magni.cs.reconstruction.gamp.config['precision_float'], np.float64)

        a_hat = self._gamp_run(
            self.z, self.F, self.F_sq, self.a, success=success)
        self.assertEqual(a_hat.dtype, np.float64)

    @unittest.skipIf(not hasattr(np, 'float128'), 'precision is not available')
    def test_float128_AMP(self, success=True):
        magni.cs.reconstruction.amp.config['precision_float'] = np.float128
        self.y = np.float128(self.y)
        self.A = np.float128(self.A)
        self.assertEqual(
            magni.cs.reconstruction.amp.config['precision_float'], np.float128)

        a_hat = self._amp_run(self.y, self.A, self.alpha, success=success)
        self.assertEqual(a_hat.dtype, np.float128)

    @unittest.skipIf(not hasattr(np, 'float128'), 'precision is not available')
    def test_float128_GAMP(self, success=True):
        input_channel_params = {'tau': self.tau, 'theta_bar': 0,
                                'theta_tilde': 1, 'use_em': False}
        output_channel_params = {'sigma_sq': 1,
                                 'noise_level_estimation': 'sample_variance'}
        sc = magni.cs.reconstruction.gamp.stop_criterion
        magni.cs.reconstruction.gamp.config.update(
            {'input_channel_parameters': input_channel_params,
             'output_channel_parameters': output_channel_params,
             'stop_criterion': sc.Residual,
             'precision_float': np.float128})
        self.z = np.float128(self.z)
        self.F = np.float128(self.F)
        self.F_sq = np.float128(self.F_sq)
        self.assertEqual(
            magni.cs.reconstruction.gamp.config['precision_float'],
            np.float128)

        a_hat = self._gamp_run(
            self.z, self.F, self.F_sq, self.a, success=success)
        self.assertEqual(a_hat.dtype, np.float128)

    @unittest.skipIf(not hasattr(np, 'float128'), 'precision is not available')
    def test_float128_GAMP_EM(self, success=True):
        input_channel_params = {'tau': self.tau, 'theta_bar': 0,
                                'theta_tilde': 1, 'use_em': True}
        output_channel_params = {'sigma_sq': 1,
                                 'noise_level_estimation': 'em'}
        sc = magni.cs.reconstruction.gamp.stop_criterion
        magni.cs.reconstruction.gamp.config.update(
            {'input_channel_parameters': input_channel_params,
             'output_channel_parameters': output_channel_params,
             'stop_criterion': sc.Residual,
             'precision_float': np.float128})
        self.z = np.float128(self.z)
        self.F = np.float128(self.F)
        self.F_sq = np.float128(self.F_sq)
        self.assertEqual(
            magni.cs.reconstruction.gamp.config['precision_float'],
            np.float128)

        a_hat = self._gamp_run(
            self.z, self.F, self.F_sq, self.a, success=success)
        self.assertEqual(a_hat.dtype, np.float128)

    def test_float32_IST(self, success=True):
        magni.cs.reconstruction.it.config.update(
            {'precision_float': np.float32})
        self.A = np.float32(self.A)
        self.y = np.float32(self.y)
        self.assertEqual(
            magni.cs.reconstruction.it.config['precision_float'], np.float32)

        alpha_hat = self._ist_run(self.y, self.A, self.alpha, success=success)
        self.assertEqual(alpha_hat.dtype, np.float32)

    def test_float64_IST(self, success=True):
        magni.cs.reconstruction.it.config.update(
            {'precision_float': np.float64})
        self.A = np.float64(self.A)
        self.y = np.float64(self.y)
        self.assertEqual(
            magni.cs.reconstruction.it.config['precision_float'], np.float64)

        alpha_hat = self._ist_run(self.y, self.A, self.alpha, success=success)
        self.assertEqual(alpha_hat.dtype, np.float64)

    @unittest.skipIf(not hasattr(np, 'float128'), 'precision is not available')
    def test_float128_IST(self, success=True):
        magni.cs.reconstruction.it.config.update(
            {'precision_float': np.float128})
        self.A = np.float128(self.A)
        self.y = np.float128(self.y)
        self.assertEqual(
            magni.cs.reconstruction.it.config['precision_float'], np.float128)

        alpha_hat = self._ist_run(self.y, self.A, self.alpha, success=success)
        self.assertEqual(alpha_hat.dtype, np.float128)


class FeatureReportHistoryTest(FeatureTest, unittest.TestCase):
    """
    Test the report history feature in reconstruction algorithms.

    The following tests are implemented:

    - *test_MSE_CONVERGENCE_AMP* (stop based on MSE)
    - *test_MAX_INTERATIONS_AMP* (stop based on max iterations)
    - *test_MSE_CONVERGENCE_GAMP* (stop based on MSE)
    - *test_MAX_INTERATIONS_GAMP* (stop based on max iterations)
    - *test_MSE_CONVERGENCE_IST* (stop based on MSE)
    - *test_MAX_INTERATIONS_IST* (stop based on max iterations)

    """

    def test_MSE_CONVERGENCE_AMP(self, success=True):
        magni.cs.reconstruction.amp.config.update(
             {'report_history': True, 'true_solution': self.alpha})

        history = self._amp_history_run(self.y, self.A, self.alpha,
                                        success=success)

        self.assertEqual(history['stop_criterion'], 'MSECONVERGENCE')
        self.assertEqual(history['stop_reason'], 'MSECONVERGENCE')
        self.assertEqual(history['stop_iteration'], 30)
        self.assertEqual(len(history['MSE']), 32)
        self.assertEqual(len(history['threshold_parameters']), 32)
        self.assertEqual(len(history['alpha_bar']), 32)
        self.assertEqual(len(history['stop_criterion_value']), 32)

    def test_MAX_ITERATION_AMP(self, success=False):
        magni.cs.reconstruction.amp.config.update(
             {'report_history': True, 'iterations': 8})

        history = self._amp_history_run(self.y, self.A, self.alpha,
                                        success=success)

        self.assertEqual(history['stop_criterion'], 'MSECONVERGENCE')
        self.assertEqual(history['stop_reason'], 'MAX_ITERATIONS')
        self.assertEqual(history['stop_iteration'], 7)
        self.assertEqual(len(history['MSE']), 1)
        self.assertEqual(len(history['threshold_parameters']), 9)
        self.assertEqual(len(history['alpha_bar']), 9)
        self.assertEqual(len(history['stop_criterion_value']), 9)

    def test_MSE_CONVERGENCE_GAMP(self, success=True):
        input_channel_params = {'tau': self.tau, 'theta_bar': 0,
                                'theta_tilde': 1, 'use_em': False}
        output_channel_params = {'sigma_sq': 1,
                                 'noise_level_estimation': 'sample_variance'}
        magni.cs.reconstruction.gamp.config.update(
            {'input_channel_parameters': input_channel_params,
             'output_channel_parameters': output_channel_params,
             'report_history': True,
             'true_solution': self.a})

        history = self._gamp_history_run(
            self.z, self.F, self.F_sq, self.a, success=success)

        self.assertEqual(history['stop_criterion'], 'MSECONVERGENCE')
        self.assertEqual(history['stop_reason'], 'MSECONVERGENCE')
        self.assertEqual(history['stop_iteration'], 10)
        self.assertEqual(len(history['MSE']), 12)
        self.assertEqual(len(history['input_channel_parameters']), 12)
        self.assertEqual(len(history['output_channel_parameters']), 12)
        self.assertEqual(len(history['alpha_bar']), 12)
        self.assertEqual(len(history['alpha_tilde']), 12)
        self.assertEqual(len(history['stop_criterion_value']), 12)

    def test_MAX_ITERATIONS_GAMP(self, success=False):
        input_channel_params = {'tau': self.tau, 'theta_bar': 0,
                                'theta_tilde': 1, 'use_em': False}
        output_channel_params = {'sigma_sq': 1,
                                 'noise_level_estimation': 'sample_variance'}
        magni.cs.reconstruction.gamp.config.update(
            {'input_channel_parameters': input_channel_params,
             'output_channel_parameters': output_channel_params,
             'report_history': True,
             'iterations': 8})

        history = self._gamp_history_run(
            self.z, self.F, self.F_sq, self.a, success=success)

        self.assertEqual(history['stop_criterion'], 'MSECONVERGENCE')
        self.assertEqual(history['stop_reason'], 'MAX_ITERATIONS')
        self.assertEqual(history['stop_iteration'], 7)
        self.assertEqual(len(history['MSE']), 1)
        self.assertEqual(len(history['output_channel_parameters']), 9)
        self.assertEqual(len(history['output_channel_parameters']), 9)
        self.assertEqual(len(history['alpha_bar']), 9)
        self.assertEqual(len(history['alpha_tilde']), 9)
        self.assertEqual(len(history['stop_criterion_value']), 9)

    def test_MSE_CONVERGENCE_IST(self, success=True):
        magni.cs.reconstruction.it.config.update(
            {'report_history': True,
             'stop_criterion': 'mse_convergence',
             'true_solution': self.alpha})

        history = self._ist_history_run(
            self.y, self.A, self.alpha, success=False)

        self.assertEqual(history['stop_criterion'], 'MSE_CONVERGENCE')
        self.assertEqual(history['stop_reason'], 'MSE_CONVERGENCE')
        self.assertEqual(history['stop_iteration'], 3)
        self.assertEqual(len(history['MSE']), 5)
        self.assertEqual(len(history['alpha']), 5)
        self.assertEqual(len(history['stop_criterion_value']), 5)

    def test_MAX_ITERATIONS_IST(self, success=False):
        magni.cs.reconstruction.it.config.update(
            {'report_history': True,
             'stop_criterion': 'mse_convergence',
             'iterations': 2})

        history = self._ist_history_run(
            self.y, self.A, self.alpha, success=False)

        self.assertEqual(history['stop_criterion'], 'MSE_CONVERGENCE')
        self.assertEqual(history['stop_reason'], 'MAX_ITERATIONS')
        self.assertEqual(history['stop_iteration'], 1)
        self.assertEqual(len(history['MSE']), 1)
        self.assertEqual(len(history['alpha']), 3)
        self.assertEqual(len(history['stop_criterion_value']), 3)


class FeatureStopCriterionTest(FeatureTest, unittest.TestCase):
    """
    Test of the stop criterion feature in reconstruction algorithms.

    The following tests are implemented:

    - *test_AMP_stop_criterion_error_handling
    - *test_residual_AMP* (stop based on residual)
    - *test_residual_measurements_ratio_AMP* (stop based on ratio of
       measurements to residual)
    - *test_normalised_MSE_convergence_AMP* (stop based on NMSE)
    - *test_GAMP_stop_criterion_error_handling
    - *test_residual_GAMP* (stop based on residual)
    - *test_residual_measurements_ratio_GAMP* (stop based on ratio of
       measurements to residual)
    - *test_normalised_MSE_convergence_GAMP* (stop based on NMSE)
    - *test_residual_IST* (stop based on residual)
    - *test_residual_measurements_ratio_IST* (stop based on ratio of
       measurements to residual)

    """

    def test_AMP_stop_criterion_error_handling(self):
        sc = magni.cs.reconstruction.amp.stop_criterion
        with self.assertRaises(TypeError):
            sc.ValidatedStopCriterion('fail')
        with self.assertRaises(TypeError):
            sc.ValidatedStopCriterion({})('fail')
        with self.assertRaises(TypeError):
            sc.NormalisedMSEConvergence('fail')
        with self.assertRaises(TypeError):
            sc.NormalisedMSEConvergence({'tolerance': 1e-3})('fail')

    def test_residual_AMP(self, success=True):
        sc = magni.cs.reconstruction.amp.stop_criterion
        magni.cs.reconstruction.amp.config.update(
            {'stop_criterion': sc.Residual})

        self._amp_run(self.y, self.A, self.alpha, success=success)

    def test_residual_measurements_ratio_AMP(self, success=True):
        sc = magni.cs.reconstruction.amp.stop_criterion
        magni.cs.reconstruction.amp.config.update(
            {'stop_criterion': sc.ResidualMeasurementsRatio})

        self._amp_run(self.y, self.A, self.alpha, success=success)

    def test_normalised_MSE_convergence_AMP(self, success=True):
        sc = magni.cs.reconstruction.amp.stop_criterion
        magni.cs.reconstruction.amp.config.update(
            {'stop_criterion': sc.NormalisedMSEConvergence})

        self._amp_run(self.y, self.A, self.alpha, success=success)

    def test_GAMP_stop_criterion_error_handling(self):
        sc = magni.cs.reconstruction.gamp.stop_criterion
        with self.assertRaises(TypeError):
            sc.ValidatedStopCriterion('fail')
        with self.assertRaises(TypeError):
            sc.ValidatedStopCriterion({})('fail')
        with self.assertRaises(TypeError):
            sc.NormalisedMSEConvergence('fail')
        with self.assertRaises(TypeError):
            sc.NormalisedMSEConvergence({'tolerance': 1e-3})('fail')

    def test_residual_GAMP(self, success=True):
        input_channel_params = {'tau': self.tau, 'theta_bar': 0,
                                'theta_tilde': 1, 'use_em': False}
        output_channel_params = {'sigma_sq': 1,
                                 'noise_level_estimation': 'sample_variance'}
        sc = magni.cs.reconstruction.gamp.stop_criterion
        magni.cs.reconstruction.gamp.config.update(
            {'input_channel_parameters': input_channel_params,
             'output_channel_parameters': output_channel_params,
             'stop_criterion': sc.Residual})

        self._gamp_run(self.z, self.F, self.F_sq, self.a, success=success)

    def test_residual_measurements_ratio_GAMP(self, success=True):
        input_channel_params = {'tau': self.tau, 'theta_bar': 0,
                                'theta_tilde': 1, 'use_em': False}
        output_channel_params = {'sigma_sq': 1,
                                 'noise_level_estimation': 'sample_variance'}
        sc = magni.cs.reconstruction.gamp.stop_criterion
        magni.cs.reconstruction.gamp.config.update(
            {'input_channel_parameters': input_channel_params,
             'output_channel_parameters': output_channel_params,
             'stop_criterion': sc.ResidualMeasurementsRatio})

        self._gamp_run(self.z, self.F, self.F_sq, self.a, success=success)

    def test_normalised_MSE_convergence_GAMP(self, success=True):
        input_channel_params = {'tau': self.tau, 'theta_bar': 0,
                                'theta_tilde': 1, 'use_em': False}
        output_channel_params = {'sigma_sq': 1,
                                 'noise_level_estimation': 'sample_variance'}
        sc = magni.cs.reconstruction.gamp.stop_criterion
        magni.cs.reconstruction.gamp.config.update(
            {'input_channel_parameters': input_channel_params,
             'output_channel_parameters': output_channel_params,
             'stop_criterion': sc.NormalisedMSEConvergence})

        self._gamp_run(self.z, self.F, self.F_sq, self.a, success=success)

    def test_residual_IST(self, success=True):
        magni.cs.reconstruction.it.config.update(
            {'stop_criterion': 'residual'})

        self._ist_run(self.y, self.A, self.alpha, success=False)

    def test_residual_measurements_ratio_IST(self, success=True):
        magni.cs.reconstruction.it.config.update(
            {'stop_criterion': 'residual'})

        self._ist_run(self.y, self.A, self.alpha, success=False)


class FeatureWarmStartTest(FeatureTest, unittest.TestCase):
    """
    Test of the warm_start feature in reconstruction algorithms.

    The following tests are implemented:

    - *test_warm_start_IT* (Iterative thresholding)
    - *test_warm_start_AMP* (Approximate Message Passing)
    - *test_warm_start_GAMP* (Generalised Approximate Message Passing)

    """

    def test_warm_start_IT(self, success_iht=True, success_ist=True):
        it_config = {'warm_start': 0.1 * np.ones(self.alpha.shape)}
        magni.cs.reconstruction.it.config.update(it_config)
        self._iht_run(self.y, self.A, self.alpha, success=success_iht)
        self._ist_run(self.y, self.A, self.alpha, success=success_ist)
        self.assertIsNotNone(
            magni.cs.reconstruction.it.config['warm_start'])

    def test_warm_start_AMP(self, success=True):
        magni.cs.reconstruction.amp.config.update(
            {'warm_start': 0.1 * np.ones(self.a.shape)})

        self._amp_run(self.y, self.A, self.alpha, success=success)

    def test_warm_start_GAMP(self, success=True):
        input_channel_params = {'tau': self.tau, 'theta_bar': 0,
                                'theta_tilde': 1, 'use_em': False}
        output_channel_params = {'sigma_sq': 1,
                                 'noise_level_estimation': 'sample_variance'}
        magni.cs.reconstruction.gamp.config.update(
            {'input_channel_parameters': input_channel_params,
             'output_channel_parameters': output_channel_params,
             'warm_start': (0.1 * np.ones(self.a.shape),
                            2 * np.ones(self.a.shape))})

        self._gamp_run(self.z, self.F, self.F_sq, self.a, success=success)


class PhaseSpaceExtremesTest(unittest.TestCase):
    """
    Tests of border case (extreme) phase space values.

    The following tests are implemented:

    - *test_basic_setup* (not extremum)
    - *test_invalid_A_and_y* (empty A and y)
    - *test_k_equals_zero*
    - *test_k_equals_m*
    - *test_m_equals_one*
    - *test_m_equals_n*
    - *test_m_and_n_equals_one*
    - *test_n_equals_one*

    """

    def setUp(self):
        self.n = 500
        self.m = 200
        self.k = 10
        self.seed = 6021

        magni.cs.reconstruction.it.config.update(iterations=200)
        magni.cs.reconstruction.gamp.config.update(iterations=200)

        # GAMP setup
        input_channel_params = {'tau': self.k/self.n, 'theta_bar': 0,
                                'theta_tilde': 1, 'use_em': False}
        output_channel_params = {'sigma_sq': 0.5,
                                 'noise_level_estimation': 'sample_variance'}
        magni.cs.reconstruction.gamp.config.update(
            {'input_channel_parameters': input_channel_params,
             'output_channel_parameters': output_channel_params})

    def tearDown(self):
        magni.cs.reconstruction.it.config.reset()
        magni.cs.reconstruction.amp.config.reset()
        magni.cs.reconstruction.gamp.config.reset()

    def test_basic_setup(self):
        y, A, alpha = use_rademacher(self.n, self.m, self.k, seed=self.seed)
        A_asq = A**2

        self._iht_run(y, A, alpha)
        self._ist_run(y, A, alpha)
        self._gamp_run(y, A, A_asq, alpha)

    def test_invalid_A_and_y(self):
        A = np.array([])
        A_asq = np.array([])
        y = np.array([])
        alpha = np.array([])
        with self.assertRaises(ValueError):
            self._iht_run(y, A, alpha)

        with self.assertRaises(ValueError):
            self._ist_run(y, A, alpha)

        with self.assertRaises(ValueError):
            self._gamp_run(y, A, A_asq, alpha)

    def test_k_equals_zero(self):
        k = 0
        y, A, alpha = use_rademacher(self.n, self.m, k, seed=self.seed)
        A_asq = A**2

        self._iht_run(y, A, alpha)
        self._ist_run(y, A, alpha)
        self._gamp_run(y, A, A_asq, alpha)

    def test_k_equals_m(self):
        k = self.m
        y, A, alpha = use_rademacher(self.n, self.m, k, seed=self.seed)
        A_asq = A**2

        self._iht_run(y, A, alpha, success=False)
        self._ist_run(y, A, alpha, success=False)
        self._gamp_run(y, A, A_asq, alpha, success=False)

    def test_m_equals_one(self):
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            m = 1
            y, A, alpha = use_rademacher(self.n, m, self.k, seed=self.seed)
            A_asq = A**2

            self._iht_run(y, A, alpha, success=False)
            self._ist_run(y, A, alpha, success=False)
            self._gamp_run(y, A, A_asq, alpha, success=False)

    def test_m_equals_n(self):
        m = self.n
        y, A, alpha = use_rademacher(self.n, m, self.k, seed=self.seed)
        A_asq = A**2

        self._iht_run(y, A, alpha)
        self._ist_run(y, A, alpha)
        self._gamp_run(y, A, A_asq, alpha)

    def test_m_and_n_equals_one(self):
        n = 1
        m = 1
        k = 1
        y, A, alpha = use_rademacher(n, m, k, seed=self.seed)
        A_asq = A**2

        self._iht_run(y, A, alpha, success=False)
        self._ist_run(y, A, alpha, success=False)
        self._gamp_run(y, A, A_asq, alpha, success=False)

    def test_n_equals_one(self):
        n = 1
        k = 1
        y, A, alpha = use_rademacher(n, self.m, k, seed=self.seed)
        A_asq = A**2

        self._iht_run(y, A, alpha)
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            self._ist_run(y, A, alpha, success=False)
        self._gamp_run(y, A, A_asq, alpha)

    def _gamp_run(self, z, F, F_sq, a, success=True):
        a_hat = magni.cs.reconstruction.gamp.run(z, F, F_sq)

        if success:
            self.assertTrue(np.allclose(a_hat, a, atol=1e-2))
        else:
            self.assertFalse(np.allclose(a_hat, a, atol=1e-2))

    def _iht_run(self, y, A, alpha, success=True):
        self.assertEqual(
            magni.cs.reconstruction.it.config['threshold_operator'], 'hard')

        alpha_hat = magni.cs.reconstruction.it.run(y, A)
        if success:
            self.assertTrue(np.allclose(alpha_hat, alpha, atol=1e-2))
        else:
            self.assertFalse(np.allclose(alpha_hat, alpha, atol=1e-2))

    def _ist_run(self, y, A, alpha, success=True):
        ist_config = {'threshold_operator': 'soft'}
        magni.cs.reconstruction.it.config.update(ist_config)
        self.assertEqual(
            magni.cs.reconstruction.it.config['threshold_operator'], 'soft')

        alpha_hat = magni.cs.reconstruction.it.run(y, A)

        if success:
            self.assertTrue(np.allclose(alpha_hat, alpha, atol=1e-2))
        else:
            self.assertFalse(np.allclose(alpha_hat, alpha, atol=1e-2))


class PhaseSpaceTest(object):
    """
    Phase space test base class.

    The following tests are implemented:

    - *test_default_IT* (default configuration)
    - *test_fixed_IT* (fixed threshold)
    - *test_adaptive_fixed_IT* (adaptive step-size, fixed threshold)
    - *test_weighted_fixed_IT* (weighted, fixed threshold)
    - *test_residual_soft_threshold_AMP* (soft threshold, residual level)
    - *test_median_soft_threshold_AMP* (soft threshold, median level)
    - *test_iidsGB_AWGN_GAMP* (s-GB)
    - *test_iidsGB_AWGN_EM_GAMP* (s-GB with EM learning)
    - *test_iidBG_AWGN_GAMP* (MMSE GAMP)
    - *test_iidBG_AWGN_EM_GAMP* (MMSE GAMP with EM learning)
    - *test_iidBG_AWGN_GAMP_rangan_sum_approx* (rangan sum approx GAMP)
    - *test_iidBG_AWGN_GAMP_krzakala_sum_approx* (krzakala sum approx GAMP)

    """

    def setUp(self, n=None, delta=None, rho=None, seed=6021):
        m = int(delta * n)

        self.k = int(rho * m)
        self.tau = delta * rho

        self.y, self.A, self.alpha = use_rademacher(n, m, self.k, seed=seed)
        self.oracle_support = self.alpha != 0

        self.z, self.F, self.a = use_gaussian(n, m, self.k, seed=seed)
        self.F_sq = self.F**2

        magni.cs.reconstruction.it.config.update(iterations=200)
        magni.cs.reconstruction.gamp.config.update(iterations=200)

    def tearDown(self):
        magni.cs.reconstruction.it.config.reset()
        magni.cs.reconstruction.amp.config.reset()
        magni.cs.reconstruction.gamp.config.reset()

    def test_residual_soft_threshold_AMP(self, success=True):
        magni.cs.reconstruction.amp.config['threshold_parameters'] = {
            'threshold_level_update_method': 'residual'}

        self._amp_run(self.y, self.A, self.alpha, success=success)

    def test_median_soft_threshold_AMP(self, success=True):
        magni.cs.reconstruction.amp.config['threshold_parameters'] = {
            'threshold_level_update_method': 'median'}

        self._amp_run(self.y, self.A, self.alpha, success=success)

    def test_iidsGB_AWGN_GAMP(self, success=True):
        input_channel_params = {'tau': self.tau, 'theta_bar': 0,
                                'theta_tilde': 1, 'use_em': False}
        output_channel_params = {'sigma_sq': 1,
                                 'noise_level_estimation': 'sample_variance'}
        IIDsGB = magni.cs.reconstruction.gamp.input_channel.IIDsGB
        magni.cs.reconstruction.gamp.config.update(
            {'input_channel_parameters': input_channel_params,
             'output_channel_parameters': output_channel_params,
             'input_channel': IIDsGB})

        self._gamp_run(self.z, self.F, self.F_sq, self.a, success=success)

    def test_iidsGB_AWGN_EM_GAMP(self, success=True):
        input_channel_params = {'tau': self.tau, 'theta_bar': 0,
                                'theta_tilde': 1, 'use_em': True,
                                'em_damping': 0.5}
        output_channel_params = {'sigma_sq': 1,
                                 'noise_level_estimation': 'em'}
        IIDsGB = magni.cs.reconstruction.gamp.input_channel.IIDsGB
        magni.cs.reconstruction.gamp.config.update(
            {'input_channel_parameters': input_channel_params,
             'output_channel_parameters': output_channel_params,
             'input_channel': IIDsGB})

        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            self._gamp_run(self.z, self.F, self.F_sq, self.a, success=success)

    def test_iidBG_AWGN_GAMP(self, success=True):
        input_channel_params = {'tau': self.tau, 'theta_bar': 0,
                                'theta_tilde': 1, 'use_em': False}
        output_channel_params = {'sigma_sq': 1,
                                 'noise_level_estimation': 'sample_variance'}
        magni.cs.reconstruction.gamp.config.update(
            {'input_channel_parameters': input_channel_params,
             'output_channel_parameters': output_channel_params})

        self._gamp_run(self.z, self.F, self.F_sq, self.a, success=success)

    def test_iidBG_AWGN_EM_GAMP(self, success=True):
        input_channel_params = {'tau': self.tau, 'theta_bar': 0,
                                'theta_tilde': 1, 'use_em': True}
        output_channel_params = {'sigma_sq': 1,
                                 'noise_level_estimation': 'em'}
        magni.cs.reconstruction.gamp.config.update(
            {'input_channel_parameters': input_channel_params,
             'output_channel_parameters': output_channel_params})

        self._gamp_run(self.z, self.F, self.F_sq, self.a, success=success)

    def test_iidBG_AWGN_GAMP_rangan_sum_approx(self, success=True):
        input_channel_params = {'tau': self.tau, 'theta_bar': 0,
                                'theta_tilde': 1, 'use_em': False}
        output_channel_params = {'sigma_sq': 1,
                                 'noise_level_estimation': 'sample_variance'}
        magni.cs.reconstruction.gamp.config.update(
            {'input_channel_parameters': input_channel_params,
             'output_channel_parameters': output_channel_params})

        self.assertEqual(
            magni.cs.reconstruction.gamp.config['sum_approximation_constant'],
            {'rangan': 1.0})
        self._gamp_run(self.z, self.F, None, self.a, success=success)

    def test_iidBG_AWGN_GAMP_krzakala_sum_approx(self, success=True):
        input_channel_params = {'tau': self.tau, 'theta_bar': 0,
                                'theta_tilde': 1, 'use_em': False}
        output_channel_params = {'sigma_sq': 1,
                                 'noise_level_estimation': 'sample_variance'}
        magni.cs.reconstruction.gamp.config.update(
            {'input_channel_parameters': input_channel_params,
             'output_channel_parameters': output_channel_params,
             'sum_approximation_constant': {'krzakala': 1.0 / self.F.shape[0]}}
        )

        self.assertEqual(
            magni.cs.reconstruction.gamp.config['sum_approximation_constant'],
            {'krzakala': 1.0 / self.F.shape[0]})
        self._gamp_run(self.z, self.F, None, self.a, success=success)

    def test_default_IT(self, success_iht=True, success_ist=True):
        self._iht_run(self.y, self.A, self.alpha, success=success_iht)
        self._ist_run(self.y, self.A, self.alpha, success=success_ist)

    def test_fixed_IT(self, success_iht=True, success_ist=True):
        it_config = {'threshold': 'fixed',
                     'threshold_fixed': self.k}
        magni.cs.reconstruction.it.config.update(it_config)
        self._iht_run(self.y, self.A, self.alpha, success=success_iht)
        self._ist_run(self.y, self.A, self.alpha, success=success_ist)
        self.assertEqual(magni.cs.reconstruction.it.config['threshold'],
                         'fixed')
        self.assertEqual(magni.cs.reconstruction.it.config['threshold_fixed'],
                         self.k)

    def test_adaptive_fixed_IT(self, success_iht=True, success_ist=True):
        it_config = {'threshold': 'fixed',
                     'threshold_fixed': self.k,
                     'kappa': 'adaptive'}
        magni.cs.reconstruction.it.config.update(it_config)
        self._iht_run(self.y, self.A, self.alpha, success=success_iht)
        self._ist_run(self.y, self.A, self.alpha, success=success_ist)
        self.assertEqual(magni.cs.reconstruction.it.config['threshold'],
                         'fixed')
        self.assertEqual(magni.cs.reconstruction.it.config['threshold_fixed'],
                         self.k)
        self.assertEqual(magni.cs.reconstruction.it.config['kappa'],
                         'adaptive')

    def test_weighted_fixed_IT(self, success_iht=True, success_ist=True):
        threshold_weights = np.linspace(
            1, 0.5, self.alpha.shape[0]).reshape(-1, 1)
        it_config = {'threshold': 'fixed',
                     'threshold_fixed': self.k,
                     'threshold_weights': threshold_weights}
        magni.cs.reconstruction.it.config.update(it_config)
        self._wiht_run(self.y, self.A, self.alpha, success=success_iht)
        self._wist_run(self.y, self.A, self.alpha, success=success_ist)
        self.assertEqual(magni.cs.reconstruction.it.config['threshold'],
                         'fixed')
        self.assertEqual(magni.cs.reconstruction.it.config['threshold_fixed'],
                         self.k)
        self.assertTrue(np.allclose(
            magni.cs.reconstruction.it.config['threshold_weights'],
            threshold_weights))

    def _amp_run(self, y, A, a, success=True):
        threshold_params = {
            'theta': magni.cs.reconstruction.amp.util.theta_mm(
                float(A.shape[0]) / A.shape[1]), 'tau_hat_sq': 1.0}
        magni.cs.reconstruction.amp.config['threshold_parameters'].update(
            threshold_params)
        a_hat = magni.cs.reconstruction.amp.run(y, A)

        if success:
            self.assertTrue(np.allclose(a_hat, a, atol=1e-2))
        else:
            self.assertFalse(np.allclose(a_hat, a, atol=1e-2))

    def _gamp_run(self, z, F, F_sq, a, success=True):
        a_hat = magni.cs.reconstruction.gamp.run(z, F, F_sq)
        if success:
            self.assertTrue(np.allclose(a_hat, a, atol=1e-2))
        else:
            self.assertFalse(np.allclose(a_hat, a, atol=1e-2))

    def _iht_run(self, y, A, alpha, success=True):
        iht_config = {'threshold_operator': 'hard'}
        magni.cs.reconstruction.it.config.update(iht_config)
        self.assertEqual(
            magni.cs.reconstruction.it.config['threshold_operator'], 'hard')

        alpha_hat = magni.cs.reconstruction.it.run(y, A)
        if success:
            self.assertTrue(np.allclose(alpha_hat, alpha, atol=1e-2))
        else:
            self.assertFalse(np.allclose(alpha_hat, alpha, atol=1e-2))

    def _ist_run(self, y, A, alpha, success=True):
        ist_config = {'threshold_operator': 'soft'}
        magni.cs.reconstruction.it.config.update(ist_config)
        self.assertEqual(
            magni.cs.reconstruction.it.config['threshold_operator'], 'soft')

        alpha_hat = magni.cs.reconstruction.it.run(y, A)

        if success:
            self.assertTrue(np.allclose(alpha_hat, alpha, atol=1e-2))
        else:
            self.assertFalse(np.allclose(alpha_hat, alpha, atol=1e-2))

    def _wiht_run(self, y, A, alpha, success=True):
        iht_config = {'threshold_operator': 'weighted_hard'}
        magni.cs.reconstruction.it.config.update(iht_config)
        self.assertEqual(
            magni.cs.reconstruction.it.config['threshold_operator'],
            'weighted_hard')

        alpha_hat = magni.cs.reconstruction.it.run(y, A)
        if success:
            self.assertTrue(np.allclose(alpha_hat, alpha, atol=1e-2))
        else:
            self.assertFalse(np.allclose(alpha_hat, alpha, atol=1e-2))

    def _wist_run(self, y, A, alpha, success=True):
        ist_config = {'threshold_operator': 'weighted_soft'}
        magni.cs.reconstruction.it.config.update(ist_config)
        self.assertEqual(
            magni.cs.reconstruction.it.config['threshold_operator'],
            'weighted_soft')

        alpha_hat = magni.cs.reconstruction.it.run(y, A)

        if success:
            self.assertTrue(np.allclose(alpha_hat, alpha, atol=1e-2))
        else:
            self.assertFalse(np.allclose(alpha_hat, alpha, atol=1e-2))


class PhaseSpaceTest1(PhaseSpaceTest, unittest.TestCase):
    """
    Test of reconstruction capabilities at Phase Space point:
    (delta, rho) = (0.08, 0.05)

    """

    def setUp(self):
        n = 500
        delta = 0.08
        rho = 0.05
        PhaseSpaceTest.setUp(self, n=n, delta=delta, rho=rho)

    def test_residual_soft_threshold_AMP(self):
        PhaseSpaceTest.test_residual_soft_threshold_AMP(self, success=False)

    def test_median_soft_threshold_AMP(self, success=True):
        PhaseSpaceTest.test_median_soft_threshold_AMP(self, success=False)

    def test_iidsGB_AWGN_GAMP(self):
        PhaseSpaceTest.test_iidsGB_AWGN_GAMP(self, success=False)

    def test_iidsGB_AWGN_EM_GAMP(self):
        PhaseSpaceTest.test_iidsGB_AWGN_EM_GAMP(self, success=False)

    def test_iidBG_AWGN_GAMP(self):
        PhaseSpaceTest.test_iidBG_AWGN_GAMP(self, success=False)

    def test_iidBG_AWGN_EM_GAMP(self):
        PhaseSpaceTest.test_iidBG_AWGN_EM_GAMP(self, success=False)

    def test_iidBG_AWGN_GAMP_rangan_sum_approx(self):
        PhaseSpaceTest.test_iidBG_AWGN_GAMP_rangan_sum_approx(
            self, success=False)

    def test_iidBG_AWGN_GAMP_krzakala_sum_approx(self):
        PhaseSpaceTest.test_iidBG_AWGN_GAMP_krzakala_sum_approx(
            self, success=False)


class PhaseSpaceTest2(PhaseSpaceTest, unittest.TestCase):
    """
    Test of reconstruction capabilities at Phase Space point:
    (delta, rho) = (0.24, 0.01)

    """

    def setUp(self):
        n = 500
        delta = 0.24
        rho = 0.01
        PhaseSpaceTest.setUp(self, n=n, delta=delta, rho=rho)

    def test_residual_soft_threshold_AMP(self):
        PhaseSpaceTest.test_residual_soft_threshold_AMP(self, success=False)

    def test_median_soft_threshold_AMP(self, success=True):
        PhaseSpaceTest.test_median_soft_threshold_AMP(self, success=False)

    def test_iidsGB_AWGN_GAMP(self):
        PhaseSpaceTest.test_iidsGB_AWGN_GAMP(self, success=False)

    def test_iidsGB_AWGN_EM_GAMP(self):
        PhaseSpaceTest.test_iidsGB_AWGN_EM_GAMP(self, success=False)

    def test_iidBG_AWGN_GAMP(self):
        PhaseSpaceTest.test_iidBG_AWGN_GAMP(self, success=False)

    def test_iidBG_AWGN_EM_GAMP(self):
        PhaseSpaceTest.test_iidBG_AWGN_EM_GAMP(self, success=False)

    def test_iidBG_AWGN_GAMP_rangan_sum_approx(self):
        PhaseSpaceTest.test_iidBG_AWGN_GAMP_rangan_sum_approx(
            self, success=False)

    def test_iidBG_AWGN_GAMP_krzakala_sum_approx(self):
        PhaseSpaceTest.test_iidBG_AWGN_GAMP_krzakala_sum_approx(
            self, success=False)


class PhaseSpaceTest3(PhaseSpaceTest, unittest.TestCase):
    """
    Test of reconstruction capabilities at Phase Space point:
    (delta, rho) = (0.38, 0.12)

    """

    def setUp(self):
        n = 500
        delta = 0.38
        rho = 0.12
        PhaseSpaceTest.setUp(self, n=n, delta=delta, rho=rho)

    def test_fixed_IT(self):
        PhaseSpaceTest.test_fixed_IT(self, success_ist=False)

    def test_adaptive_fixed_IT(self):
        PhaseSpaceTest.test_adaptive_fixed_IT(self, success_ist=False)


class PhaseSpaceTest4(PhaseSpaceTest, unittest.TestCase):
    """
    Test of reconstruction capabilities at Phase Space point:
    (delta, rho) = (0.62, 0.38)

    """

    def setUp(self):
        n = 500
        delta = 0.62
        rho = 0.38
        PhaseSpaceTest.setUp(self, n=n, delta=delta, rho=rho)

    def test_residual_soft_threshold_AMP(self):
        PhaseSpaceTest.test_residual_soft_threshold_AMP(self, success=False)

    def test_median_soft_threshold_AMP(self, success=True):
        PhaseSpaceTest.test_median_soft_threshold_AMP(self, success=False)

    def test_default_IT(self):
        PhaseSpaceTest.test_default_IT(self, success_iht=False,
                                       success_ist=False)

    def test_fixed_IT(self):
        PhaseSpaceTest.test_fixed_IT(self, success_iht=False,
                                     success_ist=False)

    def test_adaptive_fixed_IT(self):
        PhaseSpaceTest.test_adaptive_fixed_IT(self, success_iht=False,
                                              success_ist=False)

    def test_weighted_fixed_IT(self):
        PhaseSpaceTest.test_weighted_fixed_IT(self, success_iht=False,
                                              success_ist=False)


class PhaseSpaceTest5(PhaseSpaceTest, unittest.TestCase):
    """
    Test of reconstruction capabilities at Phase Space point:
    (delta, rho) = (0.78, 0.22)

    """

    def setUp(self):
        n = 500
        delta = 0.78
        rho = 0.22
        PhaseSpaceTest.setUp(self, n=n, delta=delta, rho=rho)

    def test_fixed_IT(self):
        PhaseSpaceTest.test_fixed_IT(self, success_ist=False)

    def test_adaptive_fixed_IT(self):
        PhaseSpaceTest.test_adaptive_fixed_IT(self, success_ist=False)

    def test_weighted_fixed_IT(self):
        PhaseSpaceTest.test_weighted_fixed_IT(self, success_ist=False)


class PhaseSpaceTest6(PhaseSpaceTest, unittest.TestCase):
    """
    Test of reconstruction capabilities at Phase Space point:
    (delta, rho) = (0.84, 0.08)

    """

    def setUp(self):
        n = 500
        delta = 0.84
        rho = 0.08
        PhaseSpaceTest.setUp(self, n=n, delta=delta, rho=rho)

    def test_fixed_IT(self):
        PhaseSpaceTest.test_fixed_IT(self, success_ist=False)

    def test_adaptive_fixed_IT(self):
        PhaseSpaceTest.test_adaptive_fixed_IT(self, success_ist=False)

    def test_weighted_fixed_IT(self):
        PhaseSpaceTest.test_weighted_fixed_IT(self, success_ist=False)


class PhaseSpaceTest7(PhaseSpaceTest, unittest.TestCase):
    """
    Test of reconstruction capabilities at Phase Space point:
    (delta, rho) = (0.96, 0.91)

    """

    def setUp(self):
        n = 500
        delta = 0.96
        rho = 0.91
        PhaseSpaceTest.setUp(self, n=n, delta=delta, rho=rho)

    def test_residual_soft_threshold_AMP(self):
        PhaseSpaceTest.test_residual_soft_threshold_AMP(self, success=False)

    def test_median_soft_threshold_AMP(self, success=True):
        PhaseSpaceTest.test_median_soft_threshold_AMP(self, success=False)

    def test_iidsGB_AWGN_GAMP(self):
        PhaseSpaceTest.test_iidsGB_AWGN_GAMP(self, success=False)

    def test_iidsGB_AWGN_EM_GAMP(self):
        PhaseSpaceTest.test_iidsGB_AWGN_EM_GAMP(self, success=False)

    def test_iidBG_AWGN_GAMP(self):
        PhaseSpaceTest.test_iidBG_AWGN_GAMP(self, success=False)

    def test_iidBG_AWGN_EM_GAMP(self):
        PhaseSpaceTest.test_iidBG_AWGN_EM_GAMP(self, success=False)

    def test_iidBG_AWGN_GAMP_rangan_sum_approx(self):
        PhaseSpaceTest.test_iidBG_AWGN_GAMP_rangan_sum_approx(
            self, success=False)

    def test_iidBG_AWGN_GAMP_krzakala_sum_approx(self):
        PhaseSpaceTest.test_iidBG_AWGN_GAMP_krzakala_sum_approx(
            self, success=False)

    def test_default_IT(self):
        PhaseSpaceTest.test_default_IT(self, success_iht=False,
                                       success_ist=False)

    def test_fixed_IT(self):
        PhaseSpaceTest.test_fixed_IT(self, success_iht=False,
                                     success_ist=False)

    def test_adaptive_fixed_IT(self):
        PhaseSpaceTest.test_adaptive_fixed_IT(self, success_iht=False,
                                              success_ist=False)

    def test_weighted_fixed_IT(self):
        PhaseSpaceTest.test_weighted_fixed_IT(self, success_iht=False,
                                              success_ist=False)


class PhaseSpaceTestA(PhaseSpaceTest, unittest.TestCase):
    """
    Test of reconstruction capabilities at Phase Space point:
    (delta, rho) = (0.06, 0.92)

    """

    def setUp(self):
        n = 500
        delta = 0.06
        rho = 0.92
        PhaseSpaceTest.setUp(self, n=n, delta=delta, rho=rho)

    def test_residual_soft_threshold_AMP(self):
        PhaseSpaceTest.test_residual_soft_threshold_AMP(self, success=False)

    def test_median_soft_threshold_AMP(self, success=True):
        PhaseSpaceTest.test_median_soft_threshold_AMP(self, success=False)

    def test_iidsGB_AWGN_GAMP(self):
        PhaseSpaceTest.test_iidsGB_AWGN_GAMP(self, success=False)

    def test_iidsGB_AWGN_EM_GAMP(self):
        PhaseSpaceTest.test_iidsGB_AWGN_EM_GAMP(self, success=False)

    def test_iidBG_AWGN_GAMP(self):
        PhaseSpaceTest.test_iidBG_AWGN_GAMP(self, success=False)

    def test_iidBG_AWGN_EM_GAMP(self):
        PhaseSpaceTest.test_iidBG_AWGN_EM_GAMP(self, success=False)

    def test_iidBG_AWGN_GAMP_rangan_sum_approx(self):
        PhaseSpaceTest.test_iidBG_AWGN_GAMP_rangan_sum_approx(
            self, success=False)

    def test_iidBG_AWGN_GAMP_krzakala_sum_approx(self):
        PhaseSpaceTest.test_iidBG_AWGN_GAMP_krzakala_sum_approx(
            self, success=False)

    def test_default_IT(self):
        PhaseSpaceTest.test_default_IT(self, success_iht=False,
                                       success_ist=False)

    def test_fixed_IT(self):
        PhaseSpaceTest.test_fixed_IT(self, success_iht=False,
                                     success_ist=False)

    def test_adaptive_fixed_IT(self):
        PhaseSpaceTest.test_adaptive_fixed_IT(self, success_iht=False,
                                              success_ist=False)

    def test_weighted_fixed_IT(self):
        PhaseSpaceTest.test_weighted_fixed_IT(self, success_iht=False,
                                              success_ist=False)


class PhaseSpaceTestB(PhaseSpaceTest, unittest.TestCase):
    """
    Test of reconstruction capabilities at Phase Space point:
    (delta, rho) = (0.19, 0.84)

    """

    def setUp(self):
        n = 500
        delta = 0.19
        rho = 0.84
        PhaseSpaceTest.setUp(self, n=n, delta=delta, rho=rho)

    def test_residual_soft_threshold_AMP(self):
        PhaseSpaceTest.test_residual_soft_threshold_AMP(self, success=False)

    def test_median_soft_threshold_AMP(self, success=True):
        PhaseSpaceTest.test_median_soft_threshold_AMP(self, success=False)

    def test_iidsGB_AWGN_GAMP(self):
        PhaseSpaceTest.test_iidsGB_AWGN_GAMP(self, success=False)

    def test_iidsGB_AWGN_EM_GAMP(self):
        PhaseSpaceTest.test_iidsGB_AWGN_EM_GAMP(self, success=False)

    def test_iidBG_AWGN_GAMP(self):
        PhaseSpaceTest.test_iidBG_AWGN_GAMP(self, success=False)

    def test_iidBG_AWGN_EM_GAMP(self):
        PhaseSpaceTest.test_iidBG_AWGN_EM_GAMP(self, success=False)

    def test_iidBG_AWGN_GAMP_rangan_sum_approx(self):
        PhaseSpaceTest.test_iidBG_AWGN_GAMP_rangan_sum_approx(
            self, success=False)

    def test_iidBG_AWGN_GAMP_krzakala_sum_approx(self):
        PhaseSpaceTest.test_iidBG_AWGN_GAMP_krzakala_sum_approx(
            self, success=False)

    def test_default_IT(self):
        PhaseSpaceTest.test_default_IT(self, success_iht=False,
                                       success_ist=False)

    def test_fixed_IT(self):
        PhaseSpaceTest.test_fixed_IT(self, success_iht=False,
                                     success_ist=False)

    def test_adaptive_fixed_IT(self):
        PhaseSpaceTest.test_adaptive_fixed_IT(self, success_iht=False,
                                              success_ist=False)

    def test_weighted_fixed_IT(self):
        PhaseSpaceTest.test_weighted_fixed_IT(self, success_iht=False,
                                              success_ist=False)


class PhaseSpaceTestC(PhaseSpaceTest, unittest.TestCase):
    """
    Test of reconstruction capabilities at Phase Space point:
    (delta, rho) = (0.29, 0.94)

    """

    def setUp(self):
        n = 500
        delta = 0.29
        rho = 0.94
        PhaseSpaceTest.setUp(self, n=n, delta=delta, rho=rho)

    def test_residual_soft_threshold_AMP(self):
        PhaseSpaceTest.test_residual_soft_threshold_AMP(self, success=False)

    def test_median_soft_threshold_AMP(self, success=True):
        PhaseSpaceTest.test_median_soft_threshold_AMP(self, success=False)

    def test_iidsGB_AWGN_GAMP(self):
        PhaseSpaceTest.test_iidsGB_AWGN_GAMP(self, success=False)

    def test_iidsGB_AWGN_EM_GAMP(self):
        PhaseSpaceTest.test_iidsGB_AWGN_EM_GAMP(self, success=False)

    def test_iidBG_AWGN_GAMP(self):
        PhaseSpaceTest.test_iidBG_AWGN_GAMP(self, success=False)

    def test_iidBG_AWGN_EM_GAMP(self):
        PhaseSpaceTest.test_iidBG_AWGN_EM_GAMP(self, success=False)

    def test_iidBG_AWGN_GAMP_rangan_sum_approx(self):
        PhaseSpaceTest.test_iidBG_AWGN_GAMP_rangan_sum_approx(
            self, success=False)

    def test_iidBG_AWGN_GAMP_krzakala_sum_approx(self):
        PhaseSpaceTest.test_iidBG_AWGN_GAMP_krzakala_sum_approx(
            self, success=False)

    def test_default_IT(self):
        PhaseSpaceTest.test_default_IT(self, success_iht=False,
                                       success_ist=False)

    def test_fixed_IT(self):
        PhaseSpaceTest.test_fixed_IT(self, success_iht=False,
                                     success_ist=False)

    def test_adaptive_fixed_IT(self):
        PhaseSpaceTest.test_adaptive_fixed_IT(self, success_iht=False,
                                              success_ist=False)

    def test_weighted_fixed_IT(self):
        PhaseSpaceTest.test_weighted_fixed_IT(self, success_iht=False,
                                              success_ist=False)


class TestUSERademacher(unittest.TestCase):
    """
    Test of the use_rademacher test fixture function.

    """

    def test_seed_6021(self):
        n = 10
        m = 6
        k = 3
        seed = 6021

        alpha_true = np.array([
            [-1], [1], [1], [0], [0], [0], [0], [0], [0], [0]])
        A_true = np.array([
            [0.3970924,  0.39094998, -0.51535881, -0.29376165,  0.80329912,
             0.2343297,  0.20381475, -0.4006275,  0.97687495, -0.02913711],
            [-0.21781685, -0.46838027, -0.39565219, -0.29879357, -0.1528902,
             -0.09484526, -0.24859693,  0.42678941, -0.17170236, -0.09260817],
            [-0.08024309, -0.24175707, -0.1299679, -0.15608146, -0.51588714,
             -0.48385891,  0.15647558, -0.54407042,  0.16007046, -0.39455782],
            [-0.11461544, -0.09242993,  0.10134369,  0.03684144,  0.24202215,
             0.22913925, -0.16115897,  0.07449874,  0.24777711, -0.20584097],
            [-0.49012155,  0.30646838,  0.27297925, -0.03009987,  0.21501576,
             -0.16483217,  0.49937075,  0.04385046,  0.26298357,  0.33893551],
            [-0.06364924,  0.68731702, -0.21930248, -0.20445363,  0.38122107,
             -0.05793133,  0.12713844, -1.14521796, -0.62776378, -0.1934683]])
        y_true = A_true.dot(alpha_true)

        y, A, alpha = use_rademacher(n, m, k, seed)

        self.assertTrue(np.alltrue(alpha_true == alpha))
        self.assertTrue(np.allclose(A_true, A))
        self.assertTrue(np.allclose(y_true, y))


class TestUSEGaussian(unittest.TestCase):
    """
    Test of the use_gaussian test fixture function.

    """

    def test_seed_6021(self):
        n = 10
        m = 6
        k = 3
        seed = 6021

        alpha_true = np.array([
            2.5616611, -0.30927792, -0.56096039, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
            0.0]).reshape(-1, 1)
        A_true = np.array([
            [0.3970924,  0.39094998, -0.51535881, -0.29376165,  0.80329912,
             0.2343297,  0.20381475, -0.4006275,  0.97687495, -0.02913711],
            [-0.21781685, -0.46838027, -0.39565219, -0.29879357, -0.1528902,
             -0.09484526, -0.24859693,  0.42678941, -0.17170236, -0.09260817],
            [-0.08024309, -0.24175707, -0.1299679, -0.15608146, -0.51588714,
             -0.48385891,  0.15647558, -0.54407042,  0.16007046, -0.39455782],
            [-0.11461544, -0.09242993,  0.10134369,  0.03684144,  0.24202215,
             0.22913925, -0.16115897,  0.07449874,  0.24777711, -0.20584097],
            [-0.49012155,  0.30646838,  0.27297925, -0.03009987,  0.21501576,
             -0.16483217,  0.49937075,  0.04385046,  0.26298357,  0.33893551],
            [-0.06364924,  0.68731702, -0.21930248, -0.20445363,  0.38122107,
             -0.05793133,  0.12713844, -1.14521796, -0.62776378, -0.1934683]])
        y_true = A_true.dot(alpha_true)

        y, A, alpha = use_gaussian(n, m, k, seed)

        self.assertTrue(np.allclose(alpha_true, alpha))
        self.assertTrue(np.allclose(A_true, A))
        self.assertTrue(np.allclose(y_true, y))


def use_gaussian(n, m, k, seed):
    """
    Prepare an instance of the USE/Gaussian problem suite

    Prepares:

    * :math:`\mathbf{A} \in \mathbb{R}^{m \times n}` from Uniform Spherical
    Ensemble (USE).
    * :math:`\mathbf{alpha} \in \mathbb{R}^{n}` with :math:`k` non-zero entries
    drawn from the standard normal distribution.
    * :math:`\mathbf{y} = \mathbf{A}\mathbf{\alpha}`

    Parameters
    ----------
    n : int
        The problem size.
    m : int
        The number of measurements.
    k : int
        The number of non-zero coefficients.
    seed : int
        The seed used in the random number generator.

    Returns
    -------
    (y, A, alpha) : tuple
        The measurements, measurement matrix, and coefficients.

    """

    @_decorate_validation
    def validate_input():
        _numeric('n', 'integer', range_='[1;inf)')
        _numeric('m', 'integer', range_='[1;inf)')
        _numeric('k', 'integer', range_='[0;inf)')
        _numeric('seed', 'integer', range_='[0;inf)')

    @_decorate_validation
    def validate_output():
        _numeric('y', ('integer', 'floating', 'complex'), shape=(m, 1))
        _numeric('A', ('integer', 'floating', 'complex'), shape=(m, n))
        _numeric('alpha', ('integer', 'floating', 'complex'), shape=(n, 1))

    validate_input()

    np.random.seed(seed=seed)

    A = 1/np.sqrt(m) * np.random.randn(m, n)
    alpha = np.zeros((n, 1))
    alpha[:k, 0] = np.random.randn(k)
    y = A.dot(alpha)

    validate_output()

    return y, A, alpha


def use_rademacher(n, m, k, seed):
    """
    Prepare an instance of the USE/Rademacher problem suite

    Prepares:

    * :math:`\mathbf{A} \in \mathbb{R}^{m \times n}` from Uniform Spherical
    Ensemble (USE).
    * :math:`\mathbf{alpha} \in \mathbb{R}^{n}` with :math:`k` non-zero entries
    drawn from the Rademacher distribution {1, -1}.
    * :math:`\mathbf{y} = \mathbf{A}\mathbf{\alpha}`

    Parameters
    ----------
    n : int
        The problem size.
    m : int
        The number of measurements.
    k : int
        The number of non-zero coefficients.
    seed : int
        The seed used in the random number generator.

    Returns
    -------
    (y, A, alpha) : tuple
        The measurements, measurement matrix, and coefficients.

    """

    @_decorate_validation
    def validate_input():
        _numeric('n', 'integer', range_='[1;inf)')
        _numeric('m', 'integer', range_='[1;inf)')
        _numeric('k', 'integer', range_='[0;inf)')
        _numeric('seed', 'integer', range_='[0;inf)')

    @_decorate_validation
    def validate_output():
        _numeric('y', ('integer', 'floating', 'complex'), shape=(m, 1))
        _numeric('A', ('integer', 'floating', 'complex'), shape=(m, n))
        _numeric('alpha', ('integer', 'floating', 'complex'), shape=(n, 1))

    validate_input()

    np.random.seed(seed=seed)

    A = 1/np.sqrt(m) * np.random.randn(m, n)
    alpha = np.zeros((n, 1))
    alpha[:k, 0] = np.random.randint(0, 2, k) * 2 - 1
    y = A.dot(alpha)

    validate_output()

    return y, A, alpha



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing public functions for the magni.imaging.measurements
subpackage.

Routine listings
----------------
plot_pattern(l, w, coords, mode, output_path=None)
    Function for visualising a scan pattern.
plot_pixel_mask(h, w, pixels, output_path=None)
    Function for visualising a pixel mask obtained from a scan pattern.

"""

from __future__ import division

import matplotlib.pyplot as plt

from magni.imaging.measurements import _util
from magni.imaging.visualisation import imshow as _imshow
from magni.utils import plotting as _plotting
from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate_generic as _generic
from magni.utils.validation import validate_numeric as _numeric


__all__ = ['plot_pattern', 'plot_pixel_mask']

_min_l = _util.min_l
_min_w = _util.min_w


def plot_pattern(l, w, coords, mode, output_path=None):
    """
    Display a plot that shows the pattern given by a set of coordinates.

    The pattern given by the `coords` is displayed on an `w` x `l` area. If
    `mode` is 'surface', `l` and `w` are regarded as measured in meters. If
    `mode` is 'image', `l` and `w` are regarded as measured in pixels. The
    `coords` are marked by filled circles and connected by straight dashed
    lines.

    Parameters
    ----------
    l : float or int
        The length/height of the area. If `mode` is 'surface', it must be a
        float. If `mode` is 'image', it must be an integer.
    w : float or int
        The width of the area. If `mode` is 'surface', it must be a
        float. If `mode` is 'image', it must be an integer.
    coords : ndarray
        The 2D array of pixels that make up the mask. Each row is a coordinate
        pair (x, y).
    mode : {'surface', 'image'}
        The display mode that dertermines the axis labeling and the type of `l`
        and `w`.
    output_path : str, optional
        Path (including file type extension) under which the plot is saved (the
        default value is None which implies, that the plot is not saved).

    Notes
    -----
    The resulting plot is displayed in a figure using `matplotlib`'s
    `pyplot.plot`.

    Examples
    --------
    For example,

    >>> import numpy as np
    >>> from magni.imaging.measurements import plot_pattern
    >>> l = 3
    >>> w = 3
    >>> coords = np.array([[0, 0], [1, 1], [2, 1]])
    >>> mode = 'image'
    >>> plot_pattern(l, w, coords, mode)

    """

    @_decorate_validation
    def validate_input():
        _generic('mode', 'string', value_in=('surface', 'image'))

        if mode == 'surface':
            _numeric('l', 'floating', range_='[{};inf)'.format(_min_l))
            _numeric('w', 'floating', range_='[{};inf)'.format(_min_w))
            _numeric('coords', 'floating', shape=(-1, 2))
            _numeric('coords[:, 0]', 'floating', range_='[0;{}]'.format(w),
                     shape=(-1,), var=coords[:, 0])
            _numeric('coords[:, 1]', 'floating', range_='[0;{}]'.format(l),
                     shape=(-1,), var=coords[:, 1])
        elif mode == 'image':
            _numeric('l', 'integer', range_='[2;inf)')
            _numeric('w', 'integer', range_='[2;inf)')
            _numeric('coords', ('integer', 'floating'), shape=(-1, 2))
            _numeric('coords[:, 0]', ('integer', 'floating'),
                     range_='[0;{})'.format(w), shape=(-1,),
                     var=coords[:, 0])
            _numeric('coords[:, 1]', ('integer', 'floating'),
                     range_='[0;{})'.format(l), shape=(-1,),
                     var=coords[:, 1])

        _generic('output_path', 'string', ignore_none=True)

    validate_input()

    figsize = plt.rcParams['figure.figsize']

    if w / l > figsize[0] / figsize[1]:
        figsize_local = [figsize[0], figsize[0] * l / w]
    else:
        figsize_local = [figsize[1] * w / l, figsize[1]]

    _plotting.setup_matplotlib({'figure': {'figsize': figsize_local}})

    fig, axes = plt.subplots(1, 1)
    axes.plot(coords[:, 0], coords[:, 1],
              marker=_plotting.markers[0],
              ls=_plotting.linestyles[1],
              ms=6)

    axes.set_xlim([0, w])
    axes.xaxis.tick_top()
    axes.invert_yaxis()
    axes.set_ylim([l, 0])
    axes.xaxis.set_label_position('top')
    axes.set_aspect('equal')

    if mode == 'surface':
        axes.set_xlabel('width [m]')
        axes.set_ylabel('length [m]')
        axes.ticklabel_format(style='sci', scilimits=(0, 0))

        # Fix position of exponential multipliers in scientific notation
        exp_texts = [axes.xaxis.get_children()[1],
                     axes.yaxis.get_children()[1]]
        vert_aligns = ['bottom', 'center']
        hori_aligns = ['center', 'right']
        xy_texts = [(0, 25), (-35, 0)]
        xys = [(1, 1), (0, 0)]
        annos = [None, None]
        for k, text in enumerate(exp_texts):
            text.set_visible(False)
            annos[k] = axes.annotate('', xys[k], xytext=xy_texts[k],
                                     xycoords='axes fraction',
                                     textcoords='offset points',
                                     va=vert_aligns[k], ha=hori_aligns[k])

        fig.canvas.mpl_connect(
            'draw_event', lambda event: annos[0].set_text(
                exp_texts[0].get_text()))
        fig.canvas.mpl_connect(
            'draw_event', lambda event: annos[1].set_text(
                exp_texts[1].get_text()))

    elif mode == 'image':
        axes.set_xlabel('width [pixels]')
        axes.set_ylabel('height [pixels]')

    plt.tight_layout()

    if output_path is not None:
        plt.savefig(output_path)

    _plotting.setup_matplotlib({'figure': {'figsize': figsize}})


def plot_pixel_mask(h, w, pixels, output_path=None):
    """
    Display a binary image that shows the given pixel mask.

    A black image with `w` x `h` pixels is created and the `pixels` are marked
    with white.

    Parameters
    ----------
    h : int
        The height of the image in pixels.
    w : int
        The width of the image in pixels.
    pixels : ndarray
        The 2D array of pixels that make up the mask. Each row is a coordinate
        pair (x, y), such that `coords` has size len(`pixels`) x 2.
    output_path : str, optional
        Path (including file type extension) under which the plot is saved (the
        default value is None which implies, that the plot is not saved).

    Notes
    -----
    The resulting image is displayed in a figure using
    `magni.imaging.visualisation.imshow`.

    Examples
    --------
    For example,

    >>> import numpy as np
    >>> from magni.imaging.measurements import plot_pixel_mask
    >>> h = 3
    >>> w = 3
    >>> pixels = np.array([[0, 0], [1, 1], [2, 1]])
    >>> plot_pixel_mask(h, w, pixels)

    """

    @_decorate_validation
    def validate_input():
        _numeric('h', 'integer', range_='[2;inf)')
        _numeric('w', 'integer', range_='[2;inf)')
        _numeric('pixels', 'integer', shape=(-1, 2))
        _numeric('pixels[:, 0]', 'integer', range_='[0;{}]'.format(w - 1),
                 shape=(-1,), var=pixels[:, 0])
        _numeric('pixels[:, 1]', 'integer', range_='[0;{}]'.format(h - 1),
                 shape=(-1,), var=pixels[:, 1])
        _generic('output_path', 'string', ignore_none=True)

    validate_input()

    mask = _util.construct_pixel_mask(h, w, pixels)

    figsize = plt.rcParams['figure.figsize']

    if w / h > figsize[0] / figsize[1]:
        figsize_local = [figsize[0], figsize[0] * h / w]
    else:
        figsize_local = [figsize[1] * w / h, figsize[1]]

    _plotting.setup_matplotlib({'figure': {'figsize': figsize_local}})

    fig, axes = plt.subplots(1, 1)
    _imshow(mask, ax=axes, show_axis='top')

    axes.set_xlabel('width [pixels]')
    axes.set_ylabel('height [pixels]')

    plt.tight_layout()

    if output_path is not None:
        plt.savefig(output_path)

    _plotting.setup_matplotlib({'figure': {'figsize': figsize}})



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Subpackage providing phase transition determination functionality.

Routine listings
----------------
config
    Configger providing configuration options for this subpackage.
io
    Module providing input/output functionality for stored phase transitions.
plotting
    Module providing plotting for this subpackage.
determine(algorithm, path, label='default', overwrite=False)
    Determine the phase transition of a reconstruction algorithm.

Notes
-----
See `_util` for documentation of `determine`.

The phase transition of a reconstruction algorithm describes the reconstruction
capabilities of that reconstruction algorithm. For a description of the concept
of phase transition, see [1]_.

References
----------
.. [1] C. S. Oxvig, P. S. Pedersen, T. Arildsen, and T. Larsen, "Surpassing the
   Theoretical 1-norm Phase Transition in Compressive Sensing by Tuning the
   Smoothed l0 Algorithm", *in IEEE International Conference on Acoustics,
   Speech and Signal Processing (ICASSP)*, Vancouver, Canada, May 26-31, 2013,
   pp. 6019-6023.

"""

from magni.cs.phase_transition._config import configger as config
from magni.cs.phase_transition import io
from magni.cs.phase_transition import plotting
from magni.cs.phase_transition._util import determine



"""
..
    Copyright (c) 2015-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing configuration options for the `magni.cs.reconstruction.gamp`
subpackage.

See also
--------
magni.cs.reconstruction._config.Configger : The Configger class used

Notes
-----
This module instantiates the `Configger` class provided by
`magni.cs.reconstruction._config.Configger`. The configuration options are the
following:

damping : float
    The damping applied to the variable side updates (the default is 0.0).
input_channel : magni.utils.validation.types.MMSEInputChannel
    The input channel to use (the default is
    magni.cs.reconstruction.gamp.input_channel.IIDBG).
input_channel_parameters : dict
    The parameters used in the input channel (no default is provided, which
    implies that this must be specified by the user).
iterations : int
    The maximum number of iterations to do (the default is 300).
output_channel : magni.utils.validation.types.MMSEOutputChannel
    The output channel to use (the default is
    magni.cs.reconstruction.gamp.output_channel.AWGN).
output_channel_parameters : dict
    The parameters used in the output channel (no default is provided, which
    implies that this must be specified by the user).
precision_float : {np.float, np.float16, np.float32, np.float64, np.float128,
    np.complex64, np.complex128, np.complex256}
    The floating point precision used for the computations (the default is
    np.float64).
report_A_asq_setup : bool
    The indicator of whether or not to print the A_asq details (the default is
    False).
report_history : bool
    The indicator of whether or not to return the progress history along with
    the result (the default is False).
stop_criterion : magni.utils.validation.types.StopCriterion
    The stop criterion to use in the iterations (the default is
    magni.cs.reconstruction.gamp.stop_criterion.MSEConvergence).
sum_approximation_constant : dict
    The method and constant used in a sum approximation of the squared system
    transform (the default is {'rangan': 1.0}, which implies that Rangan's
    uniform variance methods with the system matrix adapted ||A||_F^2/(m*n)
    constant is used).
tolerance : float
    The least acceptable stop criterion tolerance to break the interations (the
    default is 1e-6).
true_solution : ndarray or None
    The true solution to allow for tracking the convergence of the algorithm in
    the artificial setup where the true solution is known a-priori (the default
    is None, which implies that no true solution tracking is used).
warm_start : list or tuple
    The collection containing the initial guess of the solution vector
    (alpha_bar) and the solution variance vector (alpha_tilde) (the default is
    None, which implies that alpha_bar is taken to be a vector of zeros and
    alpha_tilde is taken to be a vector of ones).

"""

from __future__ import division

import numpy as np

from magni.cs.reconstruction._config import Configger as _Configger
from magni.cs.reconstruction.gamp.stop_criterion import (
    MSEConvergence as _MSEConvergence)
from magni.cs.reconstruction.gamp.input_channel import IIDBG as _IIDBG
from magni.cs.reconstruction.gamp.output_channel import AWGN as _AWGN
from magni.utils.validation import validate_generic as _generic
from magni.utils.validation import validate_levels as _levels
from magni.utils.validation import validate_numeric as _numeric
from magni.utils.validation.types import (
    MMSEInputChannel as _MMSEInputChannel,
    MMSEOutputChannel as _MMSEOutputChannel, StopCriterion as _StopCriterion)


configger = _Configger(
    {'damping': 0.0,
     'input_channel': _IIDBG,
     'input_channel_parameters': dict(),
     'iterations': 300,
     'output_channel': _AWGN,
     'output_channel_parameters': dict(),
     'precision_float': np.float64,
     'report_A_asq_setup': False,
     'report_history': False,
     'stop_criterion': _MSEConvergence,
     'sum_approximation_constant': {'rangan': 1.0},
     'tolerance': 1e-6,
     'true_solution': None,
     'warm_start': None},
    {'damping': _numeric(None, 'floating', range_='[0;1)'),
     'input_channel': _generic(None, 'class', superclass=_MMSEInputChannel),
     'input_channel_parameters': _generic(None, 'mapping'),
     'iterations': _numeric(None, 'integer', range_='[1;inf)'),
     'output_channel': _generic(None, 'class', superclass=_MMSEOutputChannel),
     'output_channel_parameters': _generic(None, 'mapping'),
     'precision_float': _generic(None, type, value_in=(
         np.float,
         getattr(np, 'float16', np.float_),
         getattr(np, 'float32', np.float_),
         getattr(np, 'float64', np.float_),
         getattr(np, 'float128', np.float_),
         getattr(np, 'complex64', np.complex_),
         getattr(np, 'complex128', np.complex_),
         getattr(np, 'complex256', np.complex_))),
     'report_A_asq_setup': _numeric(None, 'boolean'),
     'report_history': _numeric(None, 'boolean'),
     'stop_criterion': _generic(None, 'class', superclass=_StopCriterion),
     'sum_approximation_constant': _levels(None, (
         _generic(None, 'mapping', keys_in=('rangan', 'krzakala'), len_=1),
         _numeric(None, ('integer', 'floating'), range_='(0;inf)'))),
     'tolerance': _numeric(None, 'floating', range_='[0;inf)'),
     'true_solution': _numeric(
         None, ('integer', 'floating', 'complex'), shape=(-1, 1),
         ignore_none=True),
     'warm_start': _levels(None, (
         _generic(None, 'explicit collection', len_=2, ignore_none=True),
         _numeric(None, ('integer', 'floating', 'complex'), shape=(-1, 1))))})



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing the public function of the magni.utils subpackage.

"""

from __future__ import division

import os

from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate_generic as _generic


def split_path(path):
    """
    Split a path into folder path, file name, and file extension.

    The returned folder path ends with a folder separation character while the
    returned file extension starts with an extension separation character. The
    function is independent of the operating system and thus of the use of
    folder separation character and extension separation character.

    Parameters
    ----------
    path : str
        The path of the file either absolute or relative to the current working
        directory.

    Returns
    -------
    path : str
        The path of the containing folder of the input path.
    name : str
        The name of the object which the input path points to.
    ext : str
        The extension of the object which the input path points to (if any).

    Examples
    --------
    Concatenate a dummy path and split it using the present function:

    >>> import os
    >>> from magni.utils._util import split_path
    >>> path = 'folder' + os.sep + 'file' + os.path.extsep + 'extension'
    >>> parts = split_path(path)
    >>> print(tuple((parts[0][-7:-1], parts[1], parts[2][1:])))
    ('folder', 'file', 'extension')

    """

    @_decorate_validation
    def validate_input():
        _generic('path', 'string')

    validate_input()

    path = os.path.realpath(str(path))
    pos = str.rfind(path, os.path.sep) + 1
    path, name = path[:pos], path[pos:]

    if os.path.extsep in name:
        pos = str.rfind(name, os.path.extsep)
        name, ext = name[:pos], name[pos:]
    else:
        ext = ''

    return (path, name, ext)



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module for wrapping the doctests embedded in the Magni source code docstrings.

"""

from __future__ import division
import contextlib
import doctest
import os
import unittest
try:
    from StringIO import StringIO  # Python 2 byte str (Python 2 only)
except ImportError:
    from io import StringIO  # Python 3 unicode str (both Py2 and Py3)

import numpy as np

import magni


class TestDoctests(unittest.TestCase):
    """
    Test doctests in Magni source code docstrings.

    This TestCase in itself runs a TestSuite of doctests. As such all doctests
    are considered a single TestCase.

    """

    def setUp(self):
        """
        Create a TestSuite of doctests

        """

        self.doctest_suite = unittest.TestSuite()
        path, name, ext = magni.utils.split_path(__file__)
        magni_base_path = path.rsplit(os.sep, 2)[0]
        exclude_patterns = ['__pycache__', 'tests']

        for dirpath, dirnames, filenames in os.walk(magni_base_path):
            for filename in filenames:
                if filename[-3:] == '.py' and not any(
                        [exclude_pattern in dirpath
                         for exclude_pattern in exclude_patterns]):
                    self.doctest_suite.addTest(
                        doctest.DocFileSuite(
                            os.path.join(dirpath, filename),
                            module_relative=False))

    def test_all_doctests(self):
        """
        Run doctests TestSuite

        """

        with contextlib.closing(StringIO()) as stream:
            testRunner = unittest.TextTestRunner(
                stream=stream, descriptions=False, verbosity=1)
            doctest_result = testRunner.run(self.doctest_suite)
            error_msg = '{} failures in doctests.\n {!s}'.format(
                len(doctest_result.failures), stream.getvalue())
            self.assertTrue(doctest_result.wasSuccessful(), msg=error_msg)



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing functionality to remove tilt in images.

Routine listings
----------------
detilt(img, mask=None, mode='plane_flatten', degree=1, return_tilt=False)
    Function to remove tilt from an image.

"""

from __future__ import division

import numpy as np

import magni.imaging
from magni.utils.validation import decorate_validation as _decorate_validation
from magni.utils.validation import validate_generic as _generic
from magni.utils.validation import validate_numeric as _numeric


def detilt(img, mask=None, mode='plane_flatten', degree=1, return_tilt=False):
    """
    Estimate the tilt in an image and return the detilted image.

    Parameters
    ----------
    img : ndarray
        The image that is to be detilted.
    mask : ndarray, optional
        Bool array of the same size as `img` indicating the pixels to use in
        detilt (the default is None, which implies, that the the entire image
        is used)
    mode : {'line_flatten', 'plane_flatten'}, optional
        The type of detilting applied (the default is plane_flatten).
    degree : int, optional
        The degree of the polynomial used in line flattening
        (the default is 1).
    return_tilt : bool, optional
        If True, the detilted image and the estimated tilt is returned (the
        default is False).

    Returns
    -------
    img_detilt : ndarray
        Detilted image.
    tilt : ndarray, optional
        The estimated tilt (image). Only returned if return_tilt is True.

    Notes
    -----
    If `mode` is line flatten, the tilt in each horizontal line of pixels in
    the image is estimated by a polynomial fit independently of all other
    lines. If `mode` is plane flatten, the tilt is estimated by fitting a plane
    to all pixels.

    If a custom `mask` is specified, only the masked (True) pixels are used in
    the estimation of the tilt.

    Examples
    --------
    For example, line flatten an image using a degree 1 polynomial

    >>> import numpy as np
    >>> from magni.imaging.preprocessing import detilt
    >>> img = np.array([[0, 2, 3], [1, 5, 7], [3, 6, 8]], dtype=np.float)
    >>> np.set_printoptions(suppress=True)
    >>> detilt(img, mode='line_flatten', degree=1)
    array([[-0.16666667,  0.33333333, -0.16666667],
           [-0.33333333,  0.66666667, -0.33333333],
           [-0.16666667,  0.33333333, -0.16666667]])

    Or plane flatten the image based on a mask and return the tilt

    >>> mask = np.array([[1, 0, 0], [1, 0, 1], [0, 1, 1]], dtype=np.bool)
    >>> im, ti = detilt(img, mask=mask, mode='plane_flatten', return_tilt=True)
    >>> np.set_printoptions(suppress=True)
    >>> im
    array([[ 0.11111111, -0.66666667, -2.44444444],
           [-0.33333333,  0.88888889,  0.11111111],
           [ 0.22222222,  0.44444444, -0.33333333]])
    >>> ti
    array([[-0.11111111,  2.66666667,  5.44444444],
           [ 1.33333333,  4.11111111,  6.88888889],
           [ 2.77777778,  5.55555556,  8.33333333]])

    """

    @_decorate_validation
    def validate_input():
        _numeric('img', 'floating', shape=(-1, -1))
        _numeric('mask', 'boolean', shape=img.shape, ignore_none=True)
        _generic('mode', 'string', value_in=('plane_flatten', 'line_flatten'))
        _numeric('degree', 'integer', range_='[1;inf)')
        _numeric('return_tilt', 'boolean')

    validate_input()

    if mode == 'line_flatten':
        tilt = _line_flatten_tilt(img, mask, degree)

    elif mode == 'plane_flatten':
        tilt = _plane_flatten_tilt(img, mask)

    if return_tilt:
        return (img - tilt, tilt)
    else:
        return img - tilt


def _line_flatten_tilt(img, mask, degree):
    """
    Estimate tilt using the line flatten method.

    Parameters
    ----------
    img : ndarray
        The image from which the tilt is estimated.
    mask : ndarray, or None
        If not None, a bool ndarray of the the shape as `img` indicating which
        pixels should be used in estimate of tilt.
    degree : int
        The degree of the polynomial in the estimated line tilt.

    Returns
    -------
    tilt : ndarray
        The estimated tilt.

    """

    m, n = img.shape
    x = np.arange(n)

    # Shapes of matrices used in the detilting:
    # vander.shape=(n, degree+1), coef.shape=(degree+1, len(m_masked))
    vander = np.fliplr(np.vander(x, degree + 1))  # [1, x, x**2, ...]

    if mask is not None:
        tilt = np.zeros_like(img)
        for l in range(m):
            if mask[l, :].sum() >= degree + 1:  # Skip if underdetermined
                coef, res, rank, s = np.linalg.lstsq(vander[mask[l, :]],
                                                     img[l, mask[l, :]])
                tilt[l, :] = vander.dot(coef).T

    else:
        coef, res, rank, s = np.linalg.lstsq(vander, img.T)
        tilt = vander.dot(coef).T

    return tilt


def _plane_flatten_tilt(img, mask):
    """
    Estimate tilt using the plane flatten method.

    Parameters
    ----------
    img : ndarray
        The image from which the tilt is estimated.
    mask : ndarray, or None
        If not None, a bool ndarray of the the shape as `img` indicating which
        pixels should be used in estimate of tilt.

    Returns
    -------
    tilt : ndarray
        The estimated tilt.

    """

    m, n = img.shape
    x = np.arange(n)
    y = np.arange(m)
    Y = np.tile(y, n)
    X = np.repeat(x, m)

    # ---------->  x-axis (second numpy axis)
    # |
    # |  image
    # |
    # v
    #
    # y-axis (first numpy axis)

    # Plane equation: by + ax + c = z
    Q = np.column_stack([Y, X, np.ones(X.shape[0])])  # [y, x, 1]
    img_as_vec = magni.imaging.mat2vec(img)  # image values corresponding to z

    if mask is not None:
        mask = magni.imaging.mat2vec(mask).ravel()
        Q_mask = Q[mask]
        img_as_vec_mask = img_as_vec[mask]
    else:
        Q_mask = Q
        img_as_vec_mask = img_as_vec

    # Least squares solve: [Y, X, 1] [b, a, c].T = img_as_vec
    coef, res, rank, s = np.linalg.lstsq(Q_mask, img_as_vec_mask)

    z = Q.dot(coef)
    tilt = magni.imaging.vec2mat(z.reshape(m * n, 1), (m, n))

    return tilt



"""
..
    Copyright (c) 2014-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing configuration options for the `magni.afm` subpackage.

See also
--------
magni.utils.config.Configger : The Configger class used

Notes
-----
This module instantiates the `Configger` class provided by
`magni.utils.config`. The configuration options are the following:

algorithm : {'amp', 'gamp', 'it', 'iht', 'sl0'}
    The compressed sensing reconstruction algorithm subpackage to use (the
    default is 'it').

"""

from __future__ import division

from magni.utils.config import Configger as _Configger
from magni.utils.validation import validate_generic as _generic


configger = _Configger(
    {'algorithm': 'it'},
    {'algorithm': _generic(
        None, 'string', value_in=('amp', 'gamp', 'iht', 'it', 'sl0'))})



"""
..
    Copyright (c) 2015-2016, Magni developers.
    All rights reserved.
    See LICENSE.rst for further information.

Module providing thresholding operators used in Iterative Thresholding
algorithms.

Routine listings
----------------
get_function_handle(method)
    Return a function handle to a given threshold operator.
threshold_hard(var)
    The hard threshold operator.
threshold_none(var)
    The "no" threshold operator.
threshold_soft(var)
    The soft threshold operator.
threshold_weighted_hard(var)
    The weighted hard threshold operator.
threshold_weighted_soft(var)
    The weighted soft threshold operator.

"""

from __future__ import division

import numpy as np


def get_function_handle(method):
    """
    Return a function handle to a given threshold operator method.

    Parameters
    ----------
    method : str
        Identifier of the threshold operator to return a handle to.

    Returns
    -------
    f_handle : function
        Handle to threshold method defined in this globals scope.

    """

    return globals()['threshold_' + method]


def threshold_hard(var):
    """
    Threshold the entries of a vector using the hard threshold.

    Parameters
    ----------
    var : dict
        Local variables used in the threshold operation.

    Notes
    -----
    This threshold operation works "in-line" on the variables in `var`. Hence,
    this function does not return anything.

    Examples
    --------
    For example, thresholding a vector of values between -1 and 1

    >>> import copy, numpy as np, magni
    >>> from magni.cs.reconstruction.it._threshold_operators import (
    ... threshold_hard)
    >>> var = {'alpha': np.linspace(-1, 1, 10), 'threshold': 0.4}
    >>> threshold_hard(copy.copy(var))
    >>> var['alpha']
    array([-1.        , -0.77777778, -0.55555556,  0.        ,  0.        ,
            0.        ,  0.        ,  0.55555556,  0.77777778,  1.        ])

    """

    var['alpha'][np.abs(var['alpha']) < var['threshold']] = 0


def threshold_none(var):
    """
    Do not threshold the entries of a vector.

    Parameters
    ----------
    var : dict
        Local variables used in the threshold operation.

    Notes
    -----
    This is a dummy threshold operation that does nothing.

    """

    return


def threshold_soft(var):
    """
    Threshold the entries of a vector using the soft threshold.

    Parameters
    ----------
    var : dict
        Local variables used in the threshold operation.

    Notes
    -----
    This threshold operation works "in-line" on the variables in `var`. Hence,
    this function does not return anything.

    Examples
    --------
    For example, thresholding a vector of values between -1 and 1

    >>> import copy, numpy as np, magni
    >>> from magni.cs.reconstruction.it._threshold_operators import (
    ... threshold_soft)
    >>> var = {'alpha': np.linspace(-1, 1, 10), 'threshold': 0.4}
    >>> threshold_soft(copy.copy(var))
    >>> var['alpha']
    array([-0.6       , -0.37777778, -0.15555556,  0.        ,  0.        ,
            0.        ,  0.        ,  0.15555556,  0.37777778,  0.6       ])

    """

    x = var['alpha']
    thres = var['threshold']
    alpha_thres = ((x - thres) * (x > thres) + (x + thres) * (x < -thres))
    var['alpha'][:] = alpha_thres[:]


def threshold_weighted_hard(var):
    """
    Threshold the entries of a vector using a weighted hard threshold.

    Parameters
    ----------
    var : dict
        Local variables used in the threshold operation.

    Notes
    -----
    This threshold operation works "in-line" on the variables in `var`. Hence,
    this function does not return anything.

    Examples
    --------
    For example, thresholding a vector of values between -1 and 1

    >>> import copy, numpy as np, magni
    >>> from magni.cs.reconstruction.it._threshold_operators import (
    ... threshold_weighted_hard)
    >>> var = {'alpha': np.linspace(-1, 1, 10), 'threshold': 0.4,
    ... 'threshold_weights': 0.7 * np.ones(10)}
    >>> threshold_weighted_hard(copy.copy(var))
    >>> var['alpha']
    array([-1.        , -0.77777778,  0.        ,  0.        ,  0.        ,
            0.        ,  0.        ,  0.        ,  0.77777778,  1.        ])

    """

    mod_var = var.copy()  # Shallow copy of variable dictionary
    mod_var['alpha'] = var['alpha'].copy()  # Copy of alphas coefficents array
    mod_var['alpha'] *= var['threshold_weights']
    threshold_hard(mod_var)
    var['alpha'][mod_var['alpha'] == 0] = 0


def threshold_weighted_soft(var):
    """
    Threshold the entries of a vector using a weighted soft threshold.

    Parameters
    ----------
    var : dict
        Local variables used in the threshold operation.

    Notes
    -----
    This threshold operation works "in-line" on the variables in `var`. Hence,
    this function does not return anything.

    Examples
    --------
    For example, thresholding a vector of values between -1 and 1

    >>> import copy, numpy as np, magni
    >>> from magni.cs.reconstruction.it._threshold_operators import (
    ... threshold_weighted_soft)
    >>> var = {'alpha': np.linspace(-1, 1, 10), 'threshold': 0.4,
    ... 'threshold_weights': 0.7 * np.ones(10)}
    >>> threshold_weighted_soft(copy.copy(var))
    >>> var['alpha']
    array([-0.42857143, -0.20634921,  0.        ,  0.        ,  0.        ,
            0.        ,  0.        ,  0.        ,  0.20634921,  0.42857143])

    """

    mod_var = var.copy()  # Shallow copy of variable dictionary
    mod_var['alpha'] = var['alpha'].copy()  # Copy of alphas coefficents array
    mod_var['alpha'] *= var['threshold_weights']
    threshold_soft(mod_var)
    mod_var['alpha'] *= 1/var['threshold_weights']
    var['alpha'][:] = mod_var['alpha'][:]
